from copy import deepcopy
from typing import List

from mlwiz.util import return_class_and_args, s2c
from mlwiz.static import *


class Grid:
    r"""
    Class that implements grid-search. It computes all possible configurations
    starting from a suitable config file.

    Args:
        configs_dict (dict): the configuration dictionary specifying the
            different configurations to try
    """

    __search_type__ = GRID_SEARCH

    def __init__(self, configs_dict: dict):
        self.configs_dict = configs_dict
        self.seed = self.configs_dict.get(SEED, None)
        self._exp_name = self.configs_dict.get(EXP_NAME)
        self.data_root = self.configs_dict[DATA_ROOT]
        self.dataset_class = self.configs_dict[DATASET_CLASS]
        self.data_loader_class, self.data_loader_args = return_class_and_args(
            self.configs_dict, DATA_LOADER, return_class_name=True
        )
        self.experiment = self.configs_dict[EXPERIMENT]
        self.higher_results_are_better = self.configs_dict[
            HIGHER_RESULTS_ARE_BETTER
        ]
        self.evaluate_every = self.configs_dict[evaluate_every]
        self.device = self.configs_dict[DEVICE]
        self.dataset_getter = self.configs_dict[DATASET_GETTER]

        # This MUST be called at the END of the init method!
        self.hparams = self._gen_configs()

    def _gen_configs(self) -> List[dict]:
        r"""
        Takes a dictionary of key:list pairs and computes all possible
        combinations.

        Returns:
            A list of al possible configurations in the form of dictionaries
        """
        configs = [
            cfg
            for cfg in self._gen_helper(
                self.configs_dict[self.__search_type__]
            )
        ]
        for cfg in configs:
            cfg.update(
                {
                    DATASET_GETTER: self.dataset_getter,
                    DATA_LOADER: self.data_loader_class,
                    DATA_LOADER_ARGS: self.data_loader_args,
                    DATASET_CLASS: self.dataset_class,
                    DATA_ROOT: self.data_root,
                    DEVICE: self.device,
                    EXPERIMENT: self.experiment,
                    HIGHER_RESULTS_ARE_BETTER: self.higher_results_are_better,
                    evaluate_every: self.evaluate_every,
                }
            )
        return configs

    def _gen_helper(self, cfgs_dict: dict) -> dict:
        """
        Helper generator that yields one possible configuration at a time.
        """
        keys = cfgs_dict.keys()
        result = {}

        if cfgs_dict == {}:
            yield {}
        else:
            configs_copy = deepcopy(cfgs_dict)  # create a copy to remove keys

            # process the "first" key
            param = list(keys)[0]
            del configs_copy[param]

            first_key_values = cfgs_dict[param]

            # BASE CASE: key is associated to an atomic value
            if (
                isinstance(first_key_values, str)
                or isinstance(first_key_values, int)
                or isinstance(first_key_values, float)
                or isinstance(first_key_values, bool)
                or first_key_values is None
            ):
                # print(f'FIRST loop {first_key_values}')
                result[param] = first_key_values

                # Can be {}, hence no loop
                if configs_copy == {}:
                    yield deepcopy(result)
                else:
                    # recursive call on the other keys
                    for nested_config in self._gen_helper(configs_copy):
                        result.update(nested_config)
                        yield deepcopy(result)

            # LIST CASE: you should call _list_helper recursively on each
            # element
            elif isinstance(first_key_values, list):
                for sub_config in self._list_helper(first_key_values):
                    result[param] = sub_config

                    # Can be {}, hence no loop
                    if configs_copy == {}:
                        yield deepcopy(result)
                    else:
                        # recursive call on the other keys
                        for nested_config in self._gen_helper(configs_copy):
                            result.update(nested_config)
                            yield deepcopy(result)

            # DICT CASE: you should recursively call _grid _gen_helper on this
            # dict
            elif isinstance(first_key_values, dict):
                for value_config in self._gen_helper(first_key_values):
                    result[param] = value_config

                    # Can be {}, hence no loop
                    if configs_copy == {}:
                        yield deepcopy(result)
                    else:
                        # recursive call on the other keys
                        for nested_config in self._gen_helper(configs_copy):
                            result.update(nested_config)
                            yield deepcopy(result)

    def _list_helper(self, values: object) -> object:
        """
        Recursively parses lists of possible options for a given
        hyper-parameter.
        """
        for value in values:
            if (
                isinstance(value, str)
                or isinstance(value, int)
                or isinstance(value, float)
                or isinstance(value, bool)
                or value is None
            ):
                yield value
            elif isinstance(value, dict):
                for cfg in self._gen_helper(value):
                    yield cfg
            elif isinstance(value, list):
                for cfg in self._list_helper(value):
                    yield cfg

    def __iter__(self):
        """
        Iterates over all hyper-parameter configurations
        """
        return iter(self.hparams)

    def __len__(self):
        """
        Computes the number of hyper-parameter configurations to try
        """
        return len(self.hparams)

    def __getitem__(self, index):
        """
        Gets a specific configuration indexed by an id
        """
        return self.hparams[index]

    @property
    def exp_name(self) -> str:
        r"""
        Computes the name of the root folder

        Returns:
             the name of the root folder as made of ``EXP-NAME_DATASET-NAME``
        """
        dataset_name = s2c(self.dataset_class).__name__
        return f"{self._exp_name}_{dataset_name}"

    @property
    def num_configs(self) -> int:
        r"""
        Computes the number of configurations to try during model selection

        Returns:
             the number of configurations
        """
        return len(self)



from mlwiz.training.event.state import State


class EventDispatcher:
    """
    Class implementing the publisher/subscribe pattern. It is used to register
    subscribers that implement the
    :class:`~training.event.handler.EventHandler` interface"""

    def __init__(self):
        self._event_handlers = []

    def _dispatch(self, event_name: str, state: State):
        """
        Triggers the callback ``event_name`` for all subscribers
        (**note: order matters!**)

        Args:
            event_name (str): the name of the callback to trigger
            state (:class:`~training.event.state.State`): object holding
                training information
        """
        for event_handler in self._event_handlers:
            try:
                callback = getattr(event_handler, event_name)
            except AttributeError:
                # This happens when the callback does not implement a
                # particular method
                # e.g., when using a particular engine with new events and
                # event handlers
                callback = None

            if callback is not None:
                callback(state)

    def register(self, event_handler):
        """
        Registers a subscriber

        Args:
            event_handler: an object implementing the
                :class:`~training.event.handler.EventHandler` interface
        """
        self._event_handlers.append(event_handler)

    def unregister(self, event_handler):
        """
        De-registers a subscriber

        Args:
            event_handler: an object implementing the
                :class:`~training.event.handler.EventHandler` interface
        """
        self._event_handlers.remove(event_handler)



from copy import deepcopy

from mlwiz.evaluation.grid import Grid
from mlwiz.util import s2c
from mlwiz.static import *


class RandomSearch(Grid):
    r"""
    Class that implements random-search. It computes all possible
    configurations starting from a suitable config file.

    Args:
        configs_dict (dict): the configuration dictionary specifying the
            different configurations to try
    """

    __search_type__ = RANDOM_SEARCH

    def __init__(self, configs_dict: dict):
        self.num_samples = configs_dict[NUM_SAMPLES]
        super().__init__(configs_dict)

    def _gen_helper(self, cfgs_dict):
        r"""
        Takes a dictionary of key:list pairs and computes all possible
        combinations.

        Returns:
            A list of al possible configurations in the form of dictionaries
        """
        keys = cfgs_dict.keys()
        param = list(keys)[0]

        for _ in range(self.num_samples):
            result = {}
            for key, values in cfgs_dict.items():
                # BASE CASE: key is associated to an atomic value
                if type(values) in [str, int, float, bool, None]:
                    result[key] = values
                # DICT CASE: call _dict_helper on this dict
                elif isinstance(values, dict):
                    result[key] = self._dict_helper(deepcopy(values))

            yield deepcopy(result)

    def _dict_helper(self, configs):
        r"""
        Recursively parses a dictionary

        Returns:
            A dictionary
        """
        if SAMPLE_METHOD in configs:
            return self._sampler_helper(configs)

        for key, values in configs.items():
            if isinstance(values, dict):
                configs[key] = self._dict_helper(configs[key])

        return configs

    def _sampler_helper(self, configs):
        r"""
        Samples possible hyperparameter(s) and returns it
        (them, in this case as a dict)

         Returns:
             A dictionary
        """
        method, args = configs[SAMPLE_METHOD], configs[ARGS]
        sampler = s2c(method)
        sample = sampler(*args)

        if isinstance(sample, dict):
            return self._dict_helper(sample)

        return sample

    def __iter__(self):
        """
        Iterates over all hyper-parameter configurations (generated just once)
        """
        if self.hparams is None:
            self.hparams = self._gen_configs()
        return iter(self.hparams)

    def __len__(self):
        """
        Computes the number of hyper-parameter configurations to try
        """
        if self.hparams is None:
            self.hparams = self._gen_configs()
        return len(self.hparams)

    def __getitem__(self, index):
        """
        Gets a specific configuration indexed by an id
        """
        if self.hparams is None:
            self.hparams = self._gen_configs()
        return self.hparams[index]



from typing import List, Optional, Tuple, Union

import torch
import torchvision
from torch import relu
from torch.nn import Linear, Conv2d, MaxPool2d

from mlwiz.model.interface import ModelInterface


class MLP(ModelInterface):
    """
    An MLP model used to test the library.
    """

    def __init__(
        self,
        dim_input_features: Union[int, Tuple[int]],
        dim_target: int,
        config: dict,
    ):
        super().__init__(
            dim_input_features,
            dim_target,
            config,
        )
        assert type(dim_input_features) == int

        dim_embedding = config["dim_embedding"]
        self.mlp = torchvision.ops.MLP(
            in_channels=dim_input_features,
            hidden_channels=[dim_embedding, dim_embedding],
        )
        self.out_layer = Linear(dim_embedding, dim_target)

    def forward(
        self, data: torch.Tensor
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        """
        Implements an MLP forward pass

        Args:
            data (torch.Tensor): a batched tensor

        Returns:
            a tuple (output, embedddings)
        """
        # for testing
        if data.shape[1] == 1:
            # MNIST dataset, let's remove channel dim and flatten
            assert len(data.shape) > 2
            data = data.squeeze(1)
            data = torch.reshape(data, (-1, self.dim_input_features))
        # --

        h = self.mlp(data)
        o = self.out_layer(h)
        return o, h


class CNN(ModelInterface):
    """
    A CNN model used to test the library.
    """

    def __init__(
        self,
        dim_input_features: Union[int, Tuple[int]],
        dim_target: int,
        config: dict,
    ):
        super().__init__(
            dim_input_features,
            dim_target,
            config,
        )
        assert type(dim_input_features) == int

        # taken from https://medium.com/@bpoyeka1/building-simple-neural-networks-nn-cnn-using-pytorch-for-mnist-dataset-31e459d17788
        self.conv1 = Conv2d(
            in_channels=1,
            out_channels=8,
            kernel_size=(3, 3),
            stride=(1, 1),
            padding=(1, 1),
        )

        self.pool = MaxPool2d(kernel_size=(2, 2), stride=(2, 2))
        self.conv2 = Conv2d(
            in_channels=8,
            out_channels=16,
            kernel_size=(3, 3),
            stride=(1, 1),
            padding=(1, 1),
        )
        # last conv out was 8 so this conv input is 8.

        self.fc1 = Linear(16 * 7 * 7, dim_target)

    def forward(
        self, data: torch.Tensor
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        """
        Implements an MLP forward pass

        Args:
            data (torch.Tensor): a batched tensor

        Returns:
            a tuple (output, node_embedddings)
        """
        h = relu(self.conv1(data))
        h = self.pool(h)
        h = relu(self.conv2(h))
        h = self.pool(h)
        h = h.reshape(h.shape[0], -1)
        o = self.fc1(h)
        return o, h



import datetime
import json
import math
import os
import random
from typing import List

import torch
import tqdm

from mlwiz.data.dataset import DatasetInterface
from mlwiz.data.provider import DataProvider
from mlwiz.model.interface import ModelInterface
from mlwiz.static import *
from mlwiz.util import return_class_and_args, s2c


def clear_screen():
    """
    Clears the CLI interface.
    """
    try:
        os.system("clear")
    except Exception as e:
        try:
            os.system("cls")
        except Exception:
            pass


class ProgressManager:
    r"""
    Class that is responsible for drawing progress bars.

    Args:
        outer_folds (int): number of external folds for model assessment
        inner_folds (int): number of internal folds for model selection
        no_configs (int): number of possible configurations in model selection
        final_runs (int): number of final runs per outer fold once the
            best model has been selected
        show (bool): whether to show the progress bar or not.
            Default is ``True``
    """

    # Possible vars of ``bar_format``:
    #       * ``l_bar, bar, r_bar``,
    #       * ``n, n_fmt, total, total_fmt``,
    #       * ``percentage, elapsed, elapsed_s``,
    #       * ``ncols, nrows, desc, unit``,
    #       * ``rate, rate_fmt, rate_noinv``,
    #       * ``rate_noinv_fmt, rate_inv, rate_inv_fmt``,
    #       * ``postfix, unit_divisor, remaining, remaining_s``

    def __init__(
        self, outer_folds, inner_folds, no_configs, final_runs, show=True
    ):
        self.ncols = 100
        self.outer_folds = outer_folds
        self.inner_folds = inner_folds
        self.no_configs = no_configs
        self.final_runs = final_runs
        self.pbars = []
        self.show = show

        if not self.show:
            return

        clear_screen()
        self.show_header()
        for i in range(self.outer_folds):
            for j in range(self.inner_folds):
                self.pbars.append(self._init_selection_pbar(i, j))

        for i in range(self.outer_folds):
            self.pbars.append(self._init_assessment_pbar(i))

        self.show_footer()

        self.times = [{} for _ in range(len(self.pbars))]

    def _init_selection_pbar(self, i: int, j: int):
        """
        Initializes the progress bar for model selection

        Args:
            i (int): the id of the outer fold (from 0 to outer folds - 1)
            j (int): the id of the inner fold (from 0 to inner folds - 1)
        """
        position = i * self.inner_folds + j
        pbar = tqdm.tqdm(
            total=self.no_configs,
            ncols=self.ncols,
            ascii=True,
            position=position,
            unit="config",
            bar_format=" {desc} {percentage:3.0f}%|"
            "{bar}|{n_fmt}/{total_fmt}{postfix}",
        )
        pbar.set_description(f"Out_{i + 1}/Inn_{j + 1}")
        mean = str(datetime.timedelta(seconds=0))
        pbar.set_postfix_str(f"(1 cfg every {mean})")
        return pbar

    def _init_assessment_pbar(self, i: int):
        """
        Initializes the progress bar for risk assessment

        Args:
            i (int): the id of the outer fold (from 0 to outer folds - 1)
        """
        position = self.outer_folds * self.inner_folds + i
        pbar = tqdm.tqdm(
            total=self.final_runs,
            ncols=self.ncols,
            ascii=True,
            position=position,
            unit="config",
            bar_format=" {desc} {percentage:3.0f}%|"
            "{bar}|{n_fmt}/{total_fmt}{postfix}",
        )
        pbar.set_description(f"Final run {i + 1}")
        mean = str(datetime.timedelta(seconds=0))
        pbar.set_postfix_str(f"(1 run every {mean})")
        return pbar

    def show_header(self):
        """
        Prints the header of the progress bar
        """
        """
        \033[F --> move cursor to the beginning of the previous line
        \033[A --> move cursor up one line
        \033[<N>A --> move cursor up N lines
        """
        print(
            f'\033[F\033[A{"*" * ((self.ncols - 21) // 2 + 1)} '
            f'Experiment Progress {"*" * ((self.ncols - 21) // 2)}\n'
        )

    def show_footer(self):
        """
        Prints the footer of the progress bar
        """
        pass  # need to work how how to print after tqdm

    def refresh(self):
        """
        Refreshes the progress bar
        """

        self.show_header()
        for i, pbar in enumerate(self.pbars):
            # When resuming, do not consider completed exp. (delta approx. < 1)
            completion_times = [
                delta
                for k, (delta, completed) in self.times[i].items()
                if completed and delta > 1
            ]

            if len(completion_times) > 0:
                min_seconds = min(completion_times)
                max_seconds = max(completion_times)
                mean_seconds = sum(completion_times) / len(completion_times)
            else:
                min_seconds = 0
                max_seconds = 0
                mean_seconds = 0

            mean_time = str(datetime.timedelta(seconds=mean_seconds)).split(
                "."
            )[0]
            min_time = str(datetime.timedelta(seconds=min_seconds)).split(".")[
                0
            ]
            max_time = str(datetime.timedelta(seconds=max_seconds)).split(".")[
                0
            ]

            pbar.set_postfix_str(
                f"min:{min_time}|avg:{mean_time}|max:{max_time}"
            )

            pbar.refresh()
        self.show_footer()

    def update_state(self, msg: dict):
        """
        Updates the state of the progress bar (different from showing it
        on screen, see :func:`refresh`) once a message is received

        Args:
            msg (dict): message with updates to be parsed
        """
        if not self.show:
            return

        try:
            type = msg.get("type")

            if type == END_CONFIG:
                outer_fold = msg.get(OUTER_FOLD)
                inner_fold = msg.get(INNER_FOLD)
                config_id = msg.get(CONFIG_ID)
                position = outer_fold * self.inner_folds + inner_fold
                elapsed = msg.get(ELAPSED)
                configs_times = self.times[position]
                # Compute delta t for a specific config
                configs_times[config_id] = (
                    elapsed,
                    True,
                )  # (time.time() - configs_times[config_id][0], True)
                # Update progress bar
                self.pbars[position].update()
                self.refresh()
            elif type == END_FINAL_RUN:
                outer_fold = msg.get(OUTER_FOLD)
                run_id = msg.get(RUN_ID)
                position = self.outer_folds * self.inner_folds + outer_fold
                elapsed = msg.get(ELAPSED)
                configs_times = self.times[position]
                # Compute delta t for a specific config
                configs_times[run_id] = (
                    elapsed,
                    True,
                )  # (time.time() - configs_times[run_id][0], True)
                # Update progress bar
                self.pbars[position].update()
                self.refresh()
            else:
                raise Exception(
                    f"Cannot parse type of message {type}, fix this."
                )

        except Exception as e:
            print(e)
            return

    def __enter__(self):
        """
        Needed when Progress Manager is used as context manager.
        Does nothing besides returning self.
        """
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Needed when Progress Manager is used as context manager.
        Closes the progress bar.
        """
        for pbar in self.pbars:
            pbar.close()


"""
Various options for random search model selection
"""


def choice(*args):
    """
    Implements a random choice between a list of values
    """
    return random.choice(args)


def uniform(*args):
    """
    Implements a uniform sampling given an interval
    """
    return random.uniform(*args)


def normal(*args):
    """
    Implements a univariate normal sampling given its parameters
    """
    return random.normalvariate(*args)


def randint(*args):
    """
    Implements a random integer sampling in an interval
    """
    return random.randint(*args)


def loguniform(*args):
    r"""
    Performs a log-uniform random selection.

    Args:
        *args: a tuple of (log min, log max, [base]) to use. Base 10 is used
            if the third argument is not available.

    Returns:
        a randomly chosen value
    """
    log_min, log_max, *base = args
    base = base[0] if len(base) > 0 else 10

    log_min = math.log(log_min) / math.log(base)
    log_max = math.log(log_max) / math.log(base)

    return base ** (random.uniform(log_min, log_max))


def retrieve_experiments(
    model_selection_folder, skip_results_not_found: bool = False
) -> List[dict]:
    """
    Once the experiments are done, retrieves the config_results.json files of
    all configurations in a specific model selection folder, and returns them
    as a list of dictionaries

    :param model_selection_folder: path to the folder of a model selection,
        that is, your_results_path/..../MODEL_SELECTION/
    :param skip_results_not_found: whether to skip an experiment if a
        `config_results.json` file has not been produced yet. Useful when
        analyzing experiments while others still run.
    :return: a list of dictionaries, one per configuration, each with an extra
        key "exp_folder" which identifies the config folder.
    """
    config_directory = os.path.join(model_selection_folder)

    if not os.path.exists(config_directory):
        raise FileNotFoundError(f"Directory not found: {config_directory}")

    folder_names = []
    for _, dirs, _ in os.walk(config_directory):
        for d in dirs:
            if "config" in d:
                folder_names.append(os.path.join(config_directory, d))
        break  # do not recursively explore subfolders

    configs = []
    for cf in folder_names:
        config_results_path = os.path.join(cf, "config_results.json")
        if not os.path.exists(config_results_path) and skip_results_not_found:
            continue

        exp_info = json.load(open(config_results_path, "rb"))
        exp_config = exp_info

        exp_config["exp_folder"] = cf
        configs.append(exp_config)

    return configs


def filter_experiments(
    config_list: List[dict], logic: bool = "AND", parameters: dict = {}
):
    """
    Filters the list of configurations returned by the method ``retrieve_experiments`` according to a dictionary.
    The dictionary contains the keys and values of the configuration files you are looking for.

    If you specify more then one key/value pair to look for, then the `logic` parameter specifies whether you want to filter
    using the AND or OR rule.

    For a key, you can specify more than one possible value you are interested in by passing a list as the value, for instance
    {'device': 'cpu', 'lr': [0.1, 0.01]}

    Args:
        config_list: The list of configuration files
        logic: if ``AND``, a configuration is selected iff all conditions are satisfied. If ``OR``, a config is selected when at least
            one of the criteria is met.
        parameters: dictionary with parameters used to filter the configurations

    Returns:
        a list of filtered configurations like the one in input
    """

    def _finditem(obj, key):
        if key in obj:
            return obj[key]

        for k, v in obj.items():
            if isinstance(v, dict):
                item = _finditem(v, key)
                if item is not None:
                    return item

        return None

    assert logic in ["AND", "OR"], "logic can only be AND/OR case sensitive"

    filtered_config_list = []

    for config in config_list:
        keep = True if logic == "AND" else False

        for k, v in parameters.items():
            cf_v = _finditem(config, k)
            assert cf_v is not None, (
                f"Key {k} not found in the " f"configuration, check your input"
            )

            if type(v) == list:
                assert len(v) > 0, (
                    f'the list of values for key "{k}" cannot be'
                    f" empty, consider removing this key"
                )

                # the user specified a list of acceptable values
                # it is sufficient that one of them is present to return True
                if cf_v in v and logic == "OR":
                    keep = True
                    break

                if cf_v not in v and logic == "AND":
                    keep = False
                    break

            else:
                if v == cf_v and logic == "OR":
                    keep = True
                    break

                if v != cf_v and logic == "AND":
                    keep = False
                    break

        if keep:
            filtered_config_list.append(config)

    return filtered_config_list


def retrieve_best_configuration(model_selection_folder) -> dict:
    """
    Once the experiments are done, retrieves the winning configuration from
    a specific model selection folder, and returns it as a dictionaries

    :param model_selection_folder: path to the folder of a model selection,
        that is, your_results_path/..../MODEL_SELECTION/
    :return: a dictionary with info about the best configuration
    """
    config_directory = os.path.join(model_selection_folder)

    if not os.path.exists(config_directory):
        raise FileNotFoundError(f"Directory not found: {config_directory}")

    best_config = json.load(
        open(os.path.join(config_directory, "winner_config.json"), "rb")
    )
    return best_config


def instantiate_dataset_from_config(config: dict) -> DatasetInterface:
    """
    Instantiate a dataset from a configuration file.

    :param config (dict): the configuration file
    :return: an instance of DatasetInterface, i.e., the dataset
    """
    data_root = config[CONFIG][DATA_ROOT]
    dataset_name = config[CONFIG][DATASET]
    dataset_class = s2c(config[CONFIG][DATASET_CLASS])

    return dataset_class(data_root, dataset_name)


def instantiate_data_provider_from_config(
    config: dict, splits_filepath: str, n_outer_folds: int, n_inner_folds: int
) -> DataProvider:
    """
    Instantiate a data provider from a configuration file.
    :param config (dict): the configuration file
    :param splits_filepath (str): the path to data splits file
    :param n_outer_folds (int): the number of outer folds
    :param n_inner_folds (int): the number of inner folds
    :return: an instance of DataProvider, i.e., the data provider
    """
    data_root = config[CONFIG][DATA_ROOT]
    dataset_name = config[CONFIG][DATASET]
    dataset_class = s2c(config[CONFIG][DATASET_CLASS])
    dataset_getter = s2c(config[CONFIG][DATASET_GETTER])
    dl_class, dl_args = return_class_and_args(config[CONFIG], DATA_LOADER)

    return dataset_getter(
        data_root=data_root,
        splits_filepath=splits_filepath,
        dataset_class=dataset_class,
        dataset_name=dataset_name,
        data_loader_class=dl_class,
        data_loader_args=dl_args,
        outer_folds=n_outer_folds,
        inner_folds=n_inner_folds,
    )


def instantiate_model_from_config(
    config: dict,
    dataset: DatasetInterface,
    config_type: str = "supervised_config",
) -> ModelInterface:
    """
    Instantiate a model from a configuration file.
    :param config (dict): the configuration file
    :param dataset (DatasetInterface): the dataset used in the experiment
    :param config_type (str): the type of model in ["supervised_config", "unsupervised_config"],
        as written on the YAML experiment configuration file. Defaults to "supervised_config"
    :return: an instance of ModelInterface, i.e., the model
    """
    config_ = config[CONFIG][config_type]

    readout_str = config_.get(READOUT)
    readout_class = s2c(readout_str) if readout_str is not None else None

    model_class = s2c(config_[MODEL])
    model = model_class(
        dataset.dim_input_features,
        dataset.dim_edge_features,
        dataset.dim_target,
        readout_class,
        config=config_,
    )

    return model


def load_checkpoint(
    checkpoint_path: str, model: ModelInterface, device: torch.device
):
    """
    Load a checkpoint from a checkpoint file into a model.
    :param checkpoint_path: the checkpoint file path
    :param model (ModelInterface): the model
    :param device (torch.device): the device, e.g, "cpu" or "cuda"
    """
    ckpt_dict = torch.load(
        checkpoint_path, map_location="cpu" if device == "cpu" else None
    )
    model_state = ckpt_dict[MODEL_STATE]

    # Needed only when moving from cpu to cuda (due to changes in config
    # file). Move all parameters to cuda.
    for param in model_state.keys():
        model_state[param] = model_state[param].to(device)

    model.load_state_dict(model_state)



from mlwiz.training.event.state import State


class EventHandler:
    r"""
    Interface that adheres to the publisher/subscribe pattern for training.
    It defines the main methods that a subscriber should implement.
    Each subscriber can make use of the :class:`~training.event.state.State`
    object that is passed to each method, so detailed knowledge about that
    object is required.

    This class defines a set of callbacks that should cover a sufficient
    number of use cases. These are meant to work closely with the
    :class:`~training.callback.engine.TrainingEngine` object, which implements
    the overall training and evaluation process. This training engine is
    fairly general to accomodate a number of situations, so we expect we
    won't need to change it much to deal with static graph problems.

    We list below some pre/post conditions for each method that depend on
    the current implementation of the main training engine
    :class:`~training.callback.engine.TrainingEngine`. These are clearly not
    strict conditions, but they can help design new training engines with
    their own publisher/subscriber patterns or create subclasses of
    :class:`~training.callback.engine.TrainingEngine`
    that require special modifications.
    """

    ON_FETCH_DATA = "on_fetch_data"
    ON_FIT_START = "on_fit_start"
    ON_FIT_END = "on_fit_end"
    ON_EPOCH_START = "on_epoch_start"
    ON_EPOCH_END = "on_epoch_end"
    ON_TRAINING_EPOCH_START = "on_training_epoch_start"
    ON_TRAINING_EPOCH_END = "on_training_epoch_end"
    ON_EVAL_EPOCH_START = "on_eval_epoch_start"
    ON_EVAL_EPOCH_END = "on_eval_epoch_end"
    ON_TRAINING_BATCH_START = "on_training_batch_start"
    ON_TRAINING_BATCH_END = "on_training_batch_end"
    ON_EVAL_BATCH_START = "on_eval_batch_start"
    ON_EVAL_BATCH_END = "on_eval_batch_end"
    ON_FORWARD = "on_forward"
    ON_BACKWARD = "on_backward"
    ON_COMPUTE_METRICS = "on_compute_metrics"

    def on_fetch_data(self, state: State):
        """
        Load the next batch of data, possibly applying some kind of
        additional pre-processing not included in the
        :mod:`~mlwiz.data.transform` package.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        Pre-condition:
            The data loader is contained in  ``state.loader_iterable``
            and the minibatch ID (i.e., a counter) is stored
            in``state.id_batch``

        Post-condition:
            The ``state`` object now has a field ``batch_input`` with the
            next batch of data
        """
        pass

    def on_fit_start(self, state: State):
        """
        Initialize an object at the beginning of the training phase,
        e.g., the internals of an optimizer, using the information
        contained in ``state``.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        Pre-condition:
            The following fields have been initialized:
              * ``state.initial_epoch``: the initial epoch from which
                to start/resume training
              * ``state.stop_training``: do/don't train the model
              * ``state.optimizer_state``: the internal state of the
                optimizer (can be ``None``)
              * ``state.scheduler_state``: the internal state of the
                scheduler (can be  ``None``)
              * ``state.best_epoch_results``: a dictionary with the best
                results computed so far (can be used when resuming training,
                either for early stopping or to keep some information about
                the last checkpoint).

        """
        pass

    def on_fit_end(self, state: State):
        """
        Training has ended, free all resources, e.g., close
        Tensorboard writers.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        """
        pass

    def on_epoch_start(self, state: State):
        """
        Initialize/reset some internal state at the start of a
        training/evaluation epoch.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        Pre-condition:
            * The following fields have been initialized:
                * ``state.epoch``: the current epoch
                * ``state.return_embeddings_embeddings``:
                    do/don't return node_embeddings for each graph
                    at the end of the epoch

        """
        pass

    def on_epoch_end(self, state: State):
        """
        Perform bookkeeping operations at the end of an epoch, e.g.,
        early stopping, plotting, etc.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        Pre-condition:
            The following fields have been initialized:
                * ``state.epoch_loss``: a dictionary containing the
                    aggregated loss value across all minibatches
                * ``state.epoch_score``: a dictionary containing the
                    aggregated score value across all minibatches

        Post-condition:
            The following fields have been initialized:
                * ``state.stop_training``: do/don't train the model
                * ``state.optimizer_state``: the internal state of the
                    optimizer (can be ``None``)
                * ``state.scheduler_state``: the internal state of the
                    scheduler (can be  ``None``)
                * ``state.best_epoch_results``: a dictionary with the best
                    results computed so far (can be used when resuming
                    training, either for early stopping or to keep some
                    information about the last checkpoint).

        """
        pass

    def on_training_epoch_start(self, state: State):
        """
        Initialize/reset some internal state at the start of a training epoch.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        Pre-condition:
            The following fields have been initialized:
              * ``state.set``: it must be set to
                :const:`~mlwiz.static.TRAINING`

        """
        pass

    def on_training_epoch_end(self, state: State):
        """
        Initialize/reset some internal state at the end of a training epoch.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        Post-condition:
            The following fields have been initialized:
              * ``state.epoch_loss``: a dictionary containing the
                aggregated loss value across all minibatches
              * ``state.epoch_score``: a dictionary containing the
                aggregated score value across all minibatches

        """
        pass

    def on_eval_epoch_start(self, state: State):
        """
        Initialize/reset some internal state at the start of an
        evaluation epoch.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        Pre-condition:
            The following fields have been initialized:
              * ``state.set``: the dataset type (can be
                :const:`~mlwiz.static.TRAINING`,
                :const:`~mlwiz.static.VALIDATION`
                or :const:`~mlwiz.static.TEST`)

        """
        pass

    def on_eval_epoch_end(self, state: State):
        """
        Initialize/reset some internal state at the end of an evaluation epoch.

        Args:
            state (:class:`~training.event.state.State`): object
                holding training information

        Post-condition:
            The following fields have been initialized:
              * ``state.epoch_loss``: a dictionary containing the aggregated
                loss value across all minibatches
              * ``state.epoch_score``: a dictionary containing the aggregated
                score value across all minibatches

        """
        pass

    def on_training_batch_start(self, state: State):
        """
        Initialize/reset some internal state before training on a new
        minibatch of data.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        Pre-condition:
            The following fields have been initialized:
              * ``state.set``: it must be set to
                :const:`~mlwiz.static.TRAINING`
              * ``state.batch_input``: the input to be fed to the model
              * ``state.batch_targets``: the ground truth values to be fed
                to the model (if any, ow a dummy value can be used)
              * ``state.batch_num_targets``: the total number of ground
                truth values in the minibatch

        """
        pass

    def on_training_batch_end(self, state: State):
        """
        Initialize/reset some internal state after training on a new
        minibatch of data.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        Pre-condition:
            The following fields have been initialized:
              * ``state.set``: it must be set to
                :const:`~mlwiz.static.TRAINING`
              * ``state.batch_loss``: a dictionary holding the loss
                of the minibatch
              * ``state.batch_loss_extra``: a dictionary containing extra
                info, e.g., intermediate loss scores etc.
              * ``state.batch_score``: a dictionary holding the score
                of the minibatch

        """
        pass

    def on_eval_batch_start(self, state: State):
        """
        Initialize/reset some internal state before evaluating on a
        new minibatch of data.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        Pre-condition:
            The following fields have been initialized:
              * ``state.set``: the dataset type (can be
                :const:`~mlwiz.static.TRAINING`,
                :const:`~mlwiz.static.VALIDATION`
                or :const:`~mlwiz.static.TEST`)
              * ``state.batch_input``: the input to be fed to the model
              * ``state.batch_targets``: the ground truth values to be
                fed to the model (if any, ow a dummy value can be used)
        """
        pass

    def on_eval_batch_end(self, state: State):
        """
        Initialize/reset some internal state after evaluating on a new
        minibatch of data.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        Pre-condition:
            The following fields have been initialized:
              * ``state.set``: the dataset type (can be
                :const:`~mlwiz.static.TRAINING`,
                :const:`~mlwiz.static.VALIDATION`
                or :const:`~mlwiz.static.TEST`)
              * ``state.batch_num_graphs``: the total number of graphs
                in the minibatch
              * ``state.batch_num_nodes``: the total number of nodes
                in the minibatch
              * ``state.batch_num_targets``: the total number of ground
                truth values in the minibatch
              * ``state.batch_loss``: a dictionary holding the loss
                of the minibatch
              * ``state.batch_loss_extra``: a dictionary containing
                extra info, e.g., intermediate loss scores etc.
              * ``state.batch_score``: a dictionary holding the score
                of the minibatch

        """
        pass

    def on_forward(self, state: State):
        """
        Feed the input data to the model.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        Pre-condition:
            The following fields have been initialized:
              * ``state.batch_input``: the input to be fed to the model
              * ``state.batch_targets``: the ground truth values to be fed
                to the model (if any, ow a dummy value can be used)

        Post-condition:
            The following fields have been initialized:
              * ``state.batch_outputs``: the output produced the model
                (a tuple of values)

        """
        pass

    def on_backward(self, state: State):
        """
        Updates the parameters of the model using loss information.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        Pre-condition:
            The following fields have been initialized:
              * ``state.batch_loss``: a dictionary holding the loss
                of the minibatch

        """
        pass

    def on_compute_metrics(self, state: State):
        """
        Computes the metrics of interest using the output and ground truth
        information obtained so far. The loss-related subscriber MUST be
        called before the score-related one

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information

        Pre-condition:
            The following fields have been initialized:
              * ``state.batch_input``: the input to be fed to the model
              * ``state.batch_targets``: the ground truth values to be fed
                to the model (if any, ow a dummy value can be used)
              * ``state.batch_outputs``: the output produced the model
                (a tuple of values)

        Post-condition:
            The following fields have been initialized:
              * ``state.batch_loss``: a dictionary holding the loss
                of the minibatch
              * ``state.batch_loss_extra``: a dictionary containing extra
                info, e.g., intermediate loss scores etc.
              * ``state.batch_score``: a dictionary holding the score
                of the minibatch

        """
        pass



import random
from typing import Tuple

import numpy as np
import torch
from sklearn.model_selection import (
    StratifiedKFold,
    StratifiedShuffleSplit,
    ShuffleSplit,
    KFold,
)

import mlwiz
from mlwiz.data.dataset import IterableDatasetInterface
from mlwiz.util import s2c


class Fold:
    r"""
    Simple class that stores training, validation, and test indices.

    Args:
        train_idxs (Union[list, tuple]): training indices
        val_idxs (Union[list, tuple]): validation indices. Default is ``None``
        test_idxs (Union[list, tuple]): test indices. Default is ``None``
    """

    def __init__(self, train_idxs, val_idxs=None, test_idxs=None):
        self.train_idxs = train_idxs
        self.val_idxs = val_idxs
        self.test_idxs = test_idxs


class InnerFold(Fold):
    r"""
    Simple extension of the Fold class that returns a dictionary with
    training and validation indices (model selection).
    """

    def todict(self) -> dict:
        r"""
        Creates a dictionary with the training/validation indices.

        Returns:
            a dict with keys ``['train', 'val']`` associated with the
            respective indices
        """
        return {"train": self.train_idxs, "val": self.val_idxs}


class OuterFold(Fold):
    r"""
    Simple extension of the Fold class that returns a dictionary with training
    and test indices (risk assessment)
    """

    def todict(self) -> dict:
        r"""
        Creates a dictionary with the training/validation/test indices.

        Returns:
            a dict with keys ``['train', 'val', 'test']``
            associated with the respective indices
        """
        return {
            "train": self.train_idxs,
            "val": self.val_idxs,
            "test": self.test_idxs,
        }


class _NoShuffleTrainTestSplit:
    r"""
    Class that implements a very simple training/test split. Can be used to
    further split training data into training and validation.

    Args:
        test_ratio: percentage of data to use for evaluation.
    """

    def __init__(self, test_ratio):
        self.test_ratio = test_ratio

    # Leave the arguments as they are. The parameter `y` is needed to
    # implement the same interface as sklearn
    def split(self, idxs, y=None):
        """
        Splits the data.

        Args:
            idxs: the indices to split according to the `test_ratio` parameter
            y: Unused argument

        Returns:
            a list of a single tuple (train indices, test/eval indices)
        """
        n_samples = len(idxs)
        n_test = int(n_samples * self.test_ratio)
        n_train = n_samples - n_test
        train_idxs = np.arange(n_train)
        test_idxs = np.arange(n_train, n_train + n_test)
        return [(train_idxs, test_idxs)]


class Splitter:
    r"""
    Class that generates and stores the data splits at dataset creation time.

    Args:
        n_outer_folds (int): number of outer folds (risk assessment).
            1 means hold-out, >1 means k-fold
        n_inner_folds (int): number of inner folds (model selection).
            1 means hold-out, >1 means k-fold
        seed (int): random seed for reproducibility (on the same machine)
        stratify (bool): whether to apply stratification or not
            (should be true for classification tasks)
        shuffle (bool): whether to apply shuffle or not
        inner_val_ratio  (float): percentage of validation set for
            hold_out model selection. Default is ``0.1``
        outer_val_ratio  (float): percentage of validation set for
            hold_out model assessment (final training runs).
            Default is ``0.1``
        test_ratio  (float): percentage of test set for
            hold_out model assessment. Default is ``0.1``
    """

    def __init__(
        self,
        n_outer_folds: int,
        n_inner_folds: int,
        seed: int,
        stratify: bool,
        shuffle: bool,
        inner_val_ratio: float = 0.1,
        outer_val_ratio: float = 0.1,
        test_ratio: float = 0.1,
    ):
        self.outer_folds = []
        self.inner_folds = []
        self._stratify = stratify
        self.shuffle = shuffle

        self.n_outer_folds = n_outer_folds
        self.n_inner_folds = n_inner_folds

        self.seed = seed
        if not shuffle and seed is not None:
            print(f"Shuffle is {shuffle}, seed set to None.")
            self.seed = None

        self.inner_val_ratio = inner_val_ratio
        self.outer_val_ratio = outer_val_ratio
        self.test_ratio = test_ratio

    @property
    def stratify(self) -> bool:
        """
        Whether the splitter has to apply target stratification or not
        """
        return self._stratify

    def get_targets(
        self, dataset: mlwiz.data.dataset.DatasetInterface
    ) -> Tuple[bool, np.ndarray]:
        r"""
        Reads the entire dataset and returns the targets.

        Args:
            dataset (:class:`~mlwiz.data.dataset.DatasetInterface`):
                the dataset

        Returns:
            a tuple of two elements. The first element is a boolean, which
            is ``True`` if target values exist or an exception has not
            been thrown. The second value holds the actual targets or ``None``,
            depending on the first boolean value.
        """
        print("Extracting targets from dataset...")
        targets = []
        for sample in dataset:
            x, y = sample

            if y is None:
                print("No target found, skipping.")
                return False, None

            if not isinstance(y, torch.Tensor):
                y = torch.Tensor([y])

            targets.append(y)

        targets = torch.cat(targets, dim=0).numpy()
        print("Done.")
        return True, targets

    @classmethod
    def load(cls, path: str):
        r"""
        Loads the data splits from disk.

        Args:
            :param path: the path of the yaml file with the splits

        Returns:
            a :class:`~mlwiz.data.splitter.Splitter` object
        """
        splits = torch.load(path)

        splitter_classname = splits.get("splitter_class", "Splitter")
        splitter_class = s2c(splitter_classname)

        splitter_args = splits.get("splitter_args")
        splitter = splitter_class(**splitter_args)

        assert splitter.n_outer_folds == len(splits["outer_folds"])
        assert splitter.n_inner_folds == len(splits["inner_folds"][0])

        for fold_data in splits["outer_folds"]:
            splitter.outer_folds.append(
                OuterFold(
                    fold_data["train"],
                    val_idxs=fold_data["val"],
                    test_idxs=fold_data["test"],
                )
            )

        for inner_split in splits["inner_folds"]:
            inner_split_data = []
            for fold_data in inner_split:
                inner_split_data.append(
                    InnerFold(fold_data["train"], val_idxs=fold_data["val"])
                )
            splitter.inner_folds.append(inner_split_data)

        return splitter

    def _get_splitter(
        self, n_splits: int, stratified: bool, eval_ratio: float
    ):
        r"""
        Instantiates the appropriate splitter to use depending on the situation

        Args:
            n_splits (int): the number of different splits to create
            stratified (bool): whether to perform stratification.
                **Works with classification tasks only!**
            eval_ratio (float): the amount of evaluation (validation/test)
                data to use in case ``n_splits==1`` (i.e., hold-out data split)

        Returns:
            a :class:`~mlwiz.data.splitter.Splitter` object
        """
        if n_splits == 1:
            if not self.shuffle:
                assert (
                    stratified is False
                ), "Stratified not implemented when shuffle is False"
                splitter = _NoShuffleTrainTestSplit(test_ratio=eval_ratio)
            else:
                if stratified:
                    splitter = StratifiedShuffleSplit(
                        n_splits, test_size=eval_ratio, random_state=self.seed
                    )
                else:
                    splitter = ShuffleSplit(
                        n_splits, test_size=eval_ratio, random_state=self.seed
                    )
        elif n_splits > 1:
            if stratified:
                splitter = StratifiedKFold(
                    n_splits, shuffle=self.shuffle, random_state=self.seed
                )
            else:
                splitter = KFold(
                    n_splits, shuffle=self.shuffle, random_state=self.seed
                )
        else:
            raise ValueError(f"'n_splits' must be >=1, got {n_splits}")

        return splitter

    def split(
        self,
        dataset: mlwiz.data.dataset.DatasetInterface,
        targets: np.ndarray = None,
    ):
        r"""
        Computes the splits and stores them in the list fields
        ``self.outer_folds`` and ``self.inner_folds``.
        IMPORTANT: calling split() sets the seed of numpy, torch, and
        random for reproducibility.

        Args:
            dataset (:class:`~mlwiz.data.dataset.DatasetInterface`):
                the Dataset object
            targets (np.ndarray]): targets used for stratification.
                Default is ``None``
        """
        if self.seed is not None:
            np.random.seed(self.seed)
            torch.manual_seed(self.seed)
            torch.cuda.manual_seed(self.seed)
            random.seed(self.seed)

        idxs = range(len(dataset))

        stratified = self.stratify
        outer_idxs = np.array(idxs)

        outer_splitter = self._get_splitter(
            n_splits=self.n_outer_folds,
            stratified=stratified,
            eval_ratio=self.test_ratio,
        )  # This is the true test (outer test)

        for train_idxs, test_idxs in outer_splitter.split(
            outer_idxs, y=targets if stratified else None
        ):
            assert set(train_idxs) == set(outer_idxs[train_idxs])
            assert set(test_idxs) == set(outer_idxs[test_idxs])

            inner_fold_splits = []
            inner_idxs = outer_idxs[
                train_idxs
            ]  # equals train_idxs because outer_idxs was ordered
            inner_targets = (
                targets[train_idxs] if targets is not None else None
            )

            inner_splitter = self._get_splitter(
                n_splits=self.n_inner_folds,
                stratified=stratified,
                eval_ratio=self.inner_val_ratio,
            )  # The inner "test" is, instead, the validation set

            for inner_train_idxs, inner_val_idxs in inner_splitter.split(
                inner_idxs, y=inner_targets if stratified else None
            ):
                inner_fold = InnerFold(
                    train_idxs=inner_idxs[inner_train_idxs].tolist(),
                    val_idxs=inner_idxs[inner_val_idxs].tolist(),
                )
                inner_fold_splits.append(inner_fold)

                # False if empty
                assert not bool(
                    set(inner_train_idxs)
                    & set(inner_val_idxs)
                    & set(test_idxs)
                )
                assert not bool(
                    set(inner_idxs[inner_train_idxs])
                    & set(inner_idxs[inner_val_idxs])
                    & set(test_idxs)
                )

            self.inner_folds.append(inner_fold_splits)

            # Obtain outer val from outer train in an holdout fashion
            outer_val_splitter = self._get_splitter(
                n_splits=1,
                stratified=stratified,
                eval_ratio=self.outer_val_ratio,
            )
            outer_train_idxs, outer_val_idxs = list(
                outer_val_splitter.split(inner_idxs, y=inner_targets)
            )[0]

            # False if empty
            assert not bool(
                set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs)
            )
            assert not bool(
                set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs)
            )
            assert not bool(
                set(inner_idxs[outer_train_idxs])
                & set(inner_idxs[outer_val_idxs])
                & set(test_idxs)
            )

            np.random.shuffle(outer_train_idxs)
            np.random.shuffle(outer_val_idxs)
            np.random.shuffle(test_idxs)
            outer_fold = OuterFold(
                train_idxs=inner_idxs[outer_train_idxs].tolist(),
                val_idxs=inner_idxs[outer_val_idxs].tolist(),
                test_idxs=outer_idxs[test_idxs].tolist(),
            )
            self.outer_folds.append(outer_fold)

    def _splitter_args(self) -> dict:
        r"""
        Returns a dict with all the splitter's arguments for subsequent
        re-loading at experiment time.

        Returns:
            a dict containing all splitter's arguments.
        """
        return {
            "n_outer_folds": self.n_outer_folds,
            "n_inner_folds": self.n_inner_folds,
            "seed": self.seed,
            "stratify": self.stratify,
            "shuffle": self.shuffle,
            "inner_val_ratio": self.inner_val_ratio,
            "outer_val_ratio": self.outer_val_ratio,
            "test_ratio": self.test_ratio,
        }

    def save(self, path: str):
        r"""
        Saves the split as a dictionary into a ``torch`` file. The arguments
        of the dictionary are
        * seed (int)
        * splitter_class (str)
        * splitter_args (dict)
        * outer_folds (list of dicts)
        * inner_folds (list of lists of dicts)

        Args:
            path (str): filepath where to save the object
        """
        print("Saving splits on disk...")

        # save split class name
        module = self.__class__.__module__
        splitter_class = self.__class__.__qualname__
        if module is not None and module != "__builtin__":
            splitter_class = module + "." + splitter_class

        savedict = {
            "seed": self.seed,
            "splitter_class": splitter_class,
            "splitter_args": self._splitter_args(),
            "outer_folds": [o.todict() for o in self.outer_folds],
            "inner_folds": [],
        }

        for (
            inner_split
        ) in self.inner_folds:  # len(self.inner_folds) == # of **outer** folds
            savedict["inner_folds"].append([i.todict() for i in inner_split])
        torch.save(savedict, path)
        print("Done.")


class SingleGraphSplitter(Splitter):
    r"""
    A splitter for a single graph dataset that randomly splits nodes into
    training/validation/test

    Args:
        n_outer_folds (int): number of outer folds (risk assessment).
            1 means hold-out, >1 means k-fold
        n_inner_folds (int): number of inner folds (model selection).
            1 means hold-out, >1 means k-fold
        seed (int): random seed for reproducibility (on the same machine)
        stratify (bool): whether to apply stratification or not
            (should be true for classification tasks)
        shuffle (bool): whether to apply shuffle or not
        inner_val_ratio  (float): percentage of validation set for
            hold_out model selection. Default is ``0.1``
        outer_val_ratio  (float): percentage of validation set for
            hold_out model assessment (final training runs).
            Default is ``0.1``
        test_ratio  (float): percentage of test set for
            hold_out model assessment. Default is ``0.1``
    """

    def split(
        self,
        dataset: mlwiz.data.dataset.DatasetInterface,
        targets: np.ndarray = None,
    ):
        r"""
        Compared with the superclass version, the only difference is that the
        range of indices spans across the number of nodes of the single
        graph taken into consideration.

        Args:
            dataset (:class:`~mlwiz.data.dataset.DatasetInterface`):
                the Dataset object
            targets (np.ndarray]): targets used for stratification.
                Default is ``None``
        """
        np.random.seed(self.seed)
        torch.manual_seed(self.seed)
        torch.cuda.manual_seed(self.seed)
        random.seed(self.seed)

        assert (
            len(dataset) == 1
        ), "This class works only with single graph datasets"
        idxs = range(dataset[0][0].x.shape[0])
        stratified = self.stratify
        outer_idxs = np.array(idxs)

        # all this code below could be reused from the original splitter
        outer_splitter = self._get_splitter(
            n_splits=self.n_outer_folds,
            stratified=stratified,
            eval_ratio=self.test_ratio,
        )  # This is the true test (outer test)

        for train_idxs, test_idxs in outer_splitter.split(
            outer_idxs, y=targets
        ):
            assert set(train_idxs) == set(outer_idxs[train_idxs])
            assert set(test_idxs) == set(outer_idxs[test_idxs])

            inner_fold_splits = []
            inner_idxs = outer_idxs[
                train_idxs
            ]  # equals train_idxs because outer_idxs was ordered
            inner_targets = (
                targets[train_idxs] if targets is not None else None
            )

            inner_splitter = self._get_splitter(
                n_splits=self.n_inner_folds,
                stratified=stratified,
                eval_ratio=self.inner_val_ratio,
            )  # The inner "test" is, instead, the validation set

            for inner_train_idxs, inner_val_idxs in inner_splitter.split(
                inner_idxs, y=inner_targets
            ):
                inner_fold = InnerFold(
                    train_idxs=inner_idxs[inner_train_idxs].tolist(),
                    val_idxs=inner_idxs[inner_val_idxs].tolist(),
                )
                inner_fold_splits.append(inner_fold)

                # False if empty
                assert not bool(
                    set(inner_train_idxs)
                    & set(inner_val_idxs)
                    & set(test_idxs)
                )
                assert not bool(
                    set(inner_idxs[inner_train_idxs])
                    & set(inner_idxs[inner_val_idxs])
                    & set(test_idxs)
                )

            self.inner_folds.append(inner_fold_splits)

            # Obtain outer val from outer train in an holdout fashion
            outer_val_splitter = self._get_splitter(
                n_splits=1,
                stratified=stratified,
                eval_ratio=self.outer_val_ratio,
            )
            outer_train_idxs, outer_val_idxs = list(
                outer_val_splitter.split(inner_idxs, y=inner_targets)
            )[0]

            # False if empty
            assert not bool(
                set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs)
            )
            assert not bool(
                set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs)
            )
            assert not bool(
                set(inner_idxs[outer_train_idxs])
                & set(inner_idxs[outer_val_idxs])
                & set(test_idxs)
            )

            np.random.shuffle(outer_train_idxs)
            np.random.shuffle(outer_val_idxs)
            np.random.shuffle(test_idxs)
            outer_fold = OuterFold(
                train_idxs=inner_idxs[outer_train_idxs].tolist(),
                val_idxs=inner_idxs[outer_val_idxs].tolist(),
                test_idxs=outer_idxs[test_idxs].tolist(),
            )
            self.outer_folds.append(outer_fold)



import argparse
import os
import sys

import yaml

from mlwiz.data.util import preprocess_data
from mlwiz.static import CONFIG_FILE_CLI_ARGUMENT, CONFIG_FILE


def get_args_dict() -> dict:
    """
    Processes CLI arguments (i.e., the config file location) and returns
    a dictionary.

    Returns:
        a dictionary with the name of the configuration file in the
        :obj:`mlwiz.static.CONFIG_FILE` field.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        CONFIG_FILE_CLI_ARGUMENT,
        dest=CONFIG_FILE,
        help="config file to parse the data",
    )
    return vars(parser.parse_args())


def main():
    """
    Launches the data preparation pipeline.
    """
    # Necessary to locate dotted paths in projects that use MLWiz
    sys.path.append(os.getcwd())

    args = get_args_dict()
    options = yaml.load(open(args[CONFIG_FILE], "r"), Loader=yaml.FullLoader)
    preprocess_data(options)



import json


class Config:
    r"""
    Simple class to manage the configuration dictionary as a Python object
    with fields.

    Args:
        config_dict (dict): the configuration dictionary
    """

    def __init__(self, config_dict: dict):
        self.config_dict = config_dict

    def __getattr__(self, attr):
        """
        Returns the item associated with the key in the dictionary
        """
        return self.config_dict[attr]

    def __getitem__(self, item):
        """
        Returns the item associated with the key in the dictionary
        """
        return self.config_dict[item]

    def __contains__(self, item):
        """
        Returns true if the dictionary contains a key, false otherwise
        """
        return item in self.config_dict

    def __len__(self):
        """
        Returns the number of keys in the dictionary
        """
        return len(self.config_dict)

    def __iter__(self):
        """
        Generates an iterable object from the dictionary
        """
        return iter(self.config_dict)

    def get(self, key: str, default: object):
        """
        Returns the key from the dictionary if present, otherwise the default
        value specified

        Args:
            key (str): the key to look up in the dictionary
            default (`object`): the default object

        Returns:
            a value from the dictionaryu
        """
        return self.config_dict.get(key, default)

    def keys(self):
        r"""
        Invokes the `keys()` method of the configuration dictionary

        Returns:
            the set of keys in the dictionary
        """
        return self.config_dict.keys()

    def items(self):
        r"""
        Invokes the `items()` method of the configuration dictionary

        Returns:
            a list of (key, value) pairs
        """
        return self.config_dict.items()

    def __str__(self):
        """
        Computes an indented json representation of the dictionary
        """
        return json.dumps(self.config_dict, sort_keys=True, indent=4)



from pydoc import locate
from typing import Tuple, Callable


def return_class_and_args(
    config: dict, key: str, return_class_name: bool = False
) -> Tuple[Callable[..., object], dict]:
    r"""
    Returns the class and arguments associated to a specific key in the
    configuration file.

    Args:
        config (dict): the configuration dictionary
        key (str): a string representing a particular class in the
            configuration dictionary
        return_class_name (bool): if ``True``, returns the class name as a
            string rather than the class object

    Returns:
        a tuple (class, dict of arguments), or (None, None) if the key
        is not present in the config dictionary
    """
    if key not in config or config[key] is None:
        return None, None
    elif isinstance(config[key], str):
        return s2c(config[key]), {}
    elif isinstance(config[key], dict):
        return (
            (
                s2c(config[key]["class_name"])
                if not return_class_name
                else config[key]["class_name"]
            ),
            config[key]["args"] if "args" in config[key] else {},
        )
    else:
        raise NotImplementedError(
            f"Parameter {key} " f"has not been formatted properly"
        )


def s2c(class_name: str) -> Callable[..., object]:
    r"""
    Converts a dotted path to the corresponding class

    Args:
         class_name (str): dotted path to class name

    Returns:
        the class to be used
    """
    result = locate(class_name)
    if result is None:
        raise ImportError(
            f"The (dotted) path '{class_name}' is unknown. "
            f"Check your configuration."
        )
    return result



import torch


class State:
    """
    Any object of this class contains training information that is handled and
    modified by a :class:`TrainingEngine` as well as by the
    `EventHandler` objects implementing callbacks

    Args:
        model (torch.nn.Module): the model
        optimizer (training.callback.optimizer.Optimizer): the optimizer
        device (str): the device on which to run computations
    """

    def __init__(self, model, optimizer, device):
        self.initial_epoch = 0
        self.epoch = self.initial_epoch
        self.model = model
        self.device = device
        self.optimizer = optimizer
        self.optimizer_state = None
        self.scheduler = None
        self.scheduler_state = None
        self.stop_training = False
        self.return_embeddings_embeddings = False
        self.set = None

    def __getitem__(self, name):
        """
        Returns the value associated with argument `name`, otherwise returns
        :obj:`None`
        """
        return getattr(self, name, None)

    def __contains__(self, name):
        """
        Returns true if state contains the field `name`, and False otherwise
        """
        return name in self.__dict__

    def update(self, **values: dict):
        """
        The method sets new attributes or updates existing ones using the
        key,value pairs in ``values``

        Args:
            values: a dictionary of key,value pairs to store in
                the global state
        """
        for name, value in values.items():
            setattr(self, name, value)



from typing import Tuple, Optional, List, Union

import torch
import torch_geometric


class ModelInterface(torch.nn.Module):
    r"""
    Provides the signature for any main model to be trained under MLWiz

    Args:
        dim_input_features (Union[int, Tuple[int]]): dimension of node features
            (according to the :class:`~mlwiz.data.dataset.DatasetInterface`
            property)
        dim_target (int): dimension of the target
            (according to the :class:`~mlwiz.data.dataset.DatasetInterface`
            property)
        config (dict): config dictionary containing all the necessary
            hyper-parameters plus additional information (if needed)
    """

    def __init__(
        self,
        dim_input_features: Union[int, Tuple[int]],
        dim_target: int,
        config: dict,
    ):
        super().__init__()
        self.dim_input_features = dim_input_features
        self.dim_target = dim_target
        self.config = config

    def forward(
        self, data: Union[torch.Tensor, torch_geometric.data.Batch]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        r"""
        Performs a forward pass over a batch of graphs

        Args:
            data: a batch of samples

        Returns:
            a tuple (model's output, [optional] node embeddings,
            [optional] additional outputs
        """
        raise NotImplementedError("You need to implement this method!")



import math
import random
from copy import deepcopy
from typing import Union, Callable

import numpy as np
import torch
import torch_geometric.loader
from torch.utils.data import Subset
from torch_geometric.loader.dataloader import Collater

import mlwiz.data.dataset
from mlwiz.data.dataset import DatasetInterface
from mlwiz.data.sampler import RandomSampler
from mlwiz.data.splitter import Splitter, SingleGraphSplitter
from mlwiz.data.util import load_dataset, single_graph_collate


def seed_worker(exp_seed, worker_id):
    r"""
    Used to set a different, but reproducible, seed for all data-retriever
    workers. Without this, all workers will retrieve the data in the same
    order (important for Iterable-style datasets).

    Args:
        exp_seed (int): base seed to be used for reproducibility
        worker_id (int): id number of the worker
    """
    np.random.seed(exp_seed + worker_id)
    random.seed(exp_seed + worker_id)
    torch.manual_seed(exp_seed + worker_id)
    torch.cuda.manual_seed(exp_seed + worker_id)


class DataProvider:
    r"""
    A DataProvider object retrieves the correct data according to the external
    and internal data splits. It can be additionally used to augment the data,
    or to create a specific type of data loader. The base class does nothing
    special, but here is where the i-th element of a dataset could be
    pre-processed before constructing the mini-batches.

    IMPORTANT: if the dataset is to be shuffled, you MUST use a
    :class:`mlwiz.data.sampler.RandomSampler` object to determine the
    permutation.

    Args:
        data_root (str): the path of the root folder in which data is stored
        splits_filepath (str): the filepath of the splits. with additional
            metadata
        dataset_class
            (Callable[...,:class:`mlwiz.data.dataset.DatasetInterface`]):
            the class of the dataset
        data_loader_class
            (Union[Callable[...,:class:`torch.utils.data.DataLoader`],
            Callable[...,:class:`torch_geometric.loader.DataLoader`]]):
            the class of the data loader to use
        data_loader_args (dict): the arguments of the data loader
        outer_folds (int): the number of outer folds for risk assessment.
            1 means hold-out, >1 means k-fold
        inner_folds (int): the number of outer folds for model selection.
            1 means hold-out, >1 means k-fold

    """

    def __init__(
        self,
        data_root: str,
        splits_filepath: str,
        dataset_class: Callable[..., mlwiz.data.dataset.DatasetInterface],
        data_loader_class: Union[
            Callable[..., torch.utils.data.DataLoader],
            Callable[..., torch_geometric.loader.DataLoader],
        ],
        data_loader_args: dict,
        outer_folds: int,
        inner_folds: int,
    ) -> object:
        self.exp_seed = None

        self.data_root = data_root
        self.dataset_class = dataset_class

        self.data_loader_class = data_loader_class
        self.data_loader_args = data_loader_args

        self.outer_folds = outer_folds
        self.inner_folds = inner_folds

        self.outer_k = None
        self.inner_k = None

        self.splits_filepath = splits_filepath
        self.splitter = None

        # use this to avoid instantiating multiple versions of the same
        # dataset when no run-time
        # specific arguments are needed
        self.dataset = None

        self.dim_input_features = None
        self.dim_target = None

    def set_exp_seed(self, seed: int):
        r"""
        Sets the experiment seed to give to the DataLoader.
        Helps with reproducibility.

        Args:
            seed (int): id of the seed
        """
        self.exp_seed = seed

    def set_outer_k(self, k: int):
        r"""
        Sets the parameter k of the `risk assessment` procedure.
        Called by the evaluation modules to load the correct data subset.

        Args:
            k (int): the id of the fold, ranging from 0 to K-1.
        """
        self.outer_k = k

    def set_inner_k(self, k):
        r"""
        Sets the parameter k of the `model selection` procedure. Called by the
        evaluation modules to load the correct subset of the data.

        Args:
            k (int): the id of the fold, ranging from 0 to K-1.
        """
        self.inner_k = k

    def _get_splitter(self) -> Splitter:
        """
        Instantiates the splitter with the parameters stored
        in the file ``self.splits_filepath``

        Returns:
            a :class:`~mlwiz.data.splitter.Splitter` object
        """
        if self.splitter is None:
            self.splitter = Splitter.load(self.splits_filepath)
        return self.splitter

    def _get_dataset(self, **kwargs: dict) -> DatasetInterface:
        """
        Instantiates the dataset. Relies on the parameters stored in
        the ``dataset_kwargs.pt`` file.

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset. Not used in the base version

        Returns:
            a :class:`~mlwiz.data.dataset.DatasetInterface` object
        """
        assert "shuffle" not in kwargs, (
            "Your implementation of _get_loader should remove `shuffle` "
            "from kwargs before calling _get_dataset"
        )
        assert "batch_size" not in kwargs, (
            "Your implementation of _get_loader should remove `batch_size` "
            "from kwargs before calling _get_dataset"
        )

        if kwargs is not None and len(kwargs) != 0:
            # we probably need to pass run-time specific parameters,
            # so load the dataset in memory again
            # an example is the subset of urls in Iterable style datasets
            dataset = load_dataset(
                self.data_root, self.dataset_class, **kwargs
            )
        else:
            if self.dataset is None:
                dataset = load_dataset(self.data_root, self.dataset_class)
                self.dataset = dataset
            else:
                dataset = self.dataset

        self.dim_input_features = dataset.dim_input_features
        self.dim_target = dataset.dim_target

        return dataset

    def _get_loader(
        self, indices: list, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Instantiates the data loader.

        Args:
            indices (sequence): Indices in the whole set selected for subset
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object
        """
        shuffle = kwargs.pop("shuffle", False)
        batch_size = kwargs.pop("batch_size", 1)

        dataset: DatasetInterface = self._get_dataset(**kwargs)
        dataset = Subset(dataset, indices)

        assert (
            self.exp_seed is not None
        ), "DataLoader's seed has not been specified! Is this a bug?"

        # no need to set worker seed in map-stye dataset, see pytorch doc about reproducibility
        # this would also cause transforms based on random sampling to behave in the same way every time
        # kwargs["worker_init_fn"] = lambda worker_id: seed_worker(
        #     worker_id, self.exp_seed
        # )

        kwargs.update(self.data_loader_args)

        if shuffle is True:
            sampler = RandomSampler(dataset)
            dataloader = self.data_loader_class(
                dataset, sampler=sampler, batch_size=batch_size, **kwargs
            )
        else:
            dataloader = self.data_loader_class(
                dataset, shuffle=False, batch_size=batch_size, **kwargs
            )

        return dataloader

    def get_inner_train(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the training set for model selection associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        indices = splitter.inner_folds[self.outer_k][self.inner_k].train_idxs
        return self._get_loader(indices, **kwargs)

    def get_inner_val(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the validation set for model selection associated with
        specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        indices = splitter.inner_folds[self.outer_k][self.inner_k].val_idxs
        return self._get_loader(indices, **kwargs)

    def get_outer_train(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the training set for risk assessment associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()

        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        return self._get_loader(train_indices, **kwargs)

    def get_outer_val(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the validation set for risk assessment associated with
        specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()
        val_indices = splitter.outer_folds[self.outer_k].val_idxs
        return self._get_loader(val_indices, **kwargs)

    def get_outer_test(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the test set for risk assessment associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()
        indices = splitter.outer_folds[self.outer_k].test_idxs
        return self._get_loader(indices, **kwargs)

    def get_dim_input_features(self) -> int:
        r"""
        Returns the number of node features of the dataset

        Returns:
            the value of the property ``dim_input_features`` in the dataset

        """
        if self.dim_input_features is None:
            raise Exception(
                "You should first initialize the dataset "
                "by creating a data loader!"
            )
        return self.dim_input_features

    def get_dim_target(self) -> int:
        r"""
        Returns the dimension of the target for the task

        Returns:
            the value of the property ``dim_target`` in the dataset

        """
        if self.dim_target is None:
            raise Exception(
                "You should first initialize the dataset "
                "by creating a data loader!"
            )
        return self.dim_target


class IterableDataProvider(DataProvider):
    r"""
    A DataProvider object that allows to fetch data from an Iterable-style
    Dataset (see :class:`mlwiz.data.dataset.IterableDatasetInterface`).
    """

    def _get_loader(
        self, indices: list, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Instantiates the data loader, passing to the dataset an additional
        `url_indices` argument with the indices to fetch. This is because
        each time this method is called with different indices a separate
        instance of the dataset is called.

        Args:
            indices (sequence): Indices in the whole set selected for subset
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object
        """
        shuffle = kwargs.pop("shuffle", False)
        batch_size = kwargs.pop("batch_size", 1)

        # we can deepcopy the dataset each time the loader is called,
        # because iterable datasets are not supposed to keep all data in memory
        dataset = deepcopy(self._get_dataset())
        dataset.subset(indices)

        if shuffle:
            dataset.shuffle_urls(True)
            dataset.shuffle_urls_elements(True)

        assert (
            self.exp_seed is not None
        ), "DataLoader's seed has not been specified! Is this a bug?"

        # Define a worker_init_fn that configures each dataset copy differently
        # this is called only when num_workers is set to a value > 0
        def worker_init_fn(worker_id: int):
            """
            Set the seeds for the worker and computes the range of
            samples ids to fetch.
            """
            worker_info = torch.utils.data.get_worker_info()
            num_workers = worker_info.num_workers
            assert num_workers > 0

            # Set the random seed
            seed_worker(worker_id, self.exp_seed)

            # Get the dataset and overall length
            dataset = (
                worker_info.dataset
            )  # the dataset copy in this worker process
            dataset_length = len(
                dataset
            )  # dynamic, already refers to the subset of urls!

            per_worker = int(
                math.ceil((dataset_length) / float(worker_info.num_workers))
            )

            start = worker_id * per_worker
            end = worker_id * per_worker + per_worker

            # configure the dataset to only process the split workload
            dataset.splice(start, end)

        kwargs.update(self.data_loader_args)

        dataloader = self.data_loader_class(
            dataset,
            sampler=None,
            collate_fn=Collater(None, None),
            worker_init_fn=worker_init_fn,
            batch_size=batch_size,
            **kwargs,
        )
        return dataloader


class SingleGraphDataProvider(DataProvider):
    r"""
    A DataProvider subclass that only works with
    :class:`mlwiz.data.splitter.SingleGraphSplitter`.

    Args:
        data_root (str): the path of the root folder in which data is stored
        splits_filepath (str): the filepath of the splits. with additional
            metadata
        dataset_class
            (Callable[...,:class:`mlwiz.data.dataset.DatasetInterface`]):
            the class of the dataset
        data_loader_class
            (Union[Callable[...,:class:`torch.utils.data.DataLoader`],
            Callable[...,:class:`torch_geometric.loader.DataLoader`]]):
            the class of the data loader to use
        data_loader_args (dict): the arguments of the data loader
        outer_folds (int): the number of outer folds for risk assessment.
            1 means hold-out, >1 means k-fold
        inner_folds (int): the number of outer folds for model selection.
            1 means hold-out, >1 means k-fold

    """

    def _get_splitter(self):
        """
        Instantiates the splitter with the parameters stored in the file
        ``self.splits_filepath``. Only works
        with `~mlwiz.data.splitter.SingleGraphSplitter`.

        Returns:
            a :class:`~mlwiz.data.splitter.Splitter` object
        """
        super()._get_splitter()  # loads splitter into self.splitter
        assert isinstance(
            self.splitter, SingleGraphSplitter
        ), "This class only works with a SingleGraphNodeSplitter splitter."
        return self.splitter

    def _get_dataset(self, **kwargs: dict) -> DatasetInterface:
        """
        Compared to superclass method, this always returns a new instance of
        the dataset, optionally passing extra arguments specified at runtime.

        Args:
            kwargs (dict): a dictionary of additional arguments to be
                passed to the dataset. Not used in the base version

        Returns:
            a :class:`~mlwiz.data.dataset.DatasetInterface` object
        """
        assert "shuffle" not in kwargs, (
            "Your implementation of _get_loader should remove `shuffle` "
            "from kwargs before calling _get_dataset"
        )
        assert "batch_size" not in kwargs, (
            "Your implementation of _get_loader should remove `batch_size` "
            "from kwargs before calling _get_dataset"
        )

        # we probably need to pass run-time specific parameters, so load the
        # dataset in memory again
        # an example is the subset of urls in Iterable style datasets
        dataset = load_dataset(self.data_root, self.dataset_class, **kwargs)

        self.dim_input_features = dataset.dim_input_features
        self.dim_target = dataset.dim_target
        return dataset

    def _get_loader(
        self, eval_indices: list, training_indices: list, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Compared to superclass method, returns a dataloader with the single
        graph augmented with additional fields. These are `training_indices`
        with the indices that refer to training nodes (usually always
        available) and `eval_indices`, which specify which are the indices
        on which to evaluate (can be validation or test).

        Args:
            indices (sequence): Indices in the whole set selected for subset
            eval_set (bool): whether or not indices refer to eval set
                (validation or test) or to training
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version
        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object
        """
        shuffle = kwargs.pop("shuffle", False)
        batch_size = kwargs.pop("batch_size", 1)

        # TODO is there a way to refactor the code to avoid that we have
        #  to duplicate the dataset for every loader just to set different
        #  training and evaluation indices?
        dataset: DatasetInterface = self._get_dataset(**kwargs)

        # single graph means there is only one sample
        dataset[0][0].training_indices = torch.tensor(training_indices)
        dataset[0][0].eval_indices = torch.tensor(eval_indices)

        assert (
            self.exp_seed is not None
        ), "DataLoader's seed has not been specified! Is this a bug?"

        # no need to set worker seed in map-stye dataset, see pytorch doc about reproducibility
        # this would also cause transforms based on random sampling to behave in the same way every time
        # kwargs["worker_init_fn"] = lambda worker_id: seed_worker(
        #     worker_id, self.exp_seed
        # )

        kwargs.update(self.data_loader_args)

        if shuffle is True:
            sampler = RandomSampler(dataset)
            dataloader = self.data_loader_class(
                dataset,
                sampler=sampler,
                batch_size=batch_size,
                collate_fn=single_graph_collate,
                **kwargs,
            )
        else:
            dataloader = self.data_loader_class(
                dataset,
                shuffle=False,
                batch_size=batch_size,
                collate_fn=single_graph_collate,
                **kwargs,
            )

        return dataloader

    def get_inner_train(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the training set for model selection associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.inner_folds[self.outer_k][
            self.inner_k
        ].train_idxs
        return self._get_loader(
            train_indices, training_indices=train_indices, **kwargs
        )

    def get_inner_val(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the validation set for model selection associated with
        specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.inner_folds[self.outer_k][
            self.inner_k
        ].train_idxs
        val_indices = splitter.inner_folds[self.outer_k][self.inner_k].val_idxs
        return self._get_loader(
            val_indices, training_indices=train_indices, **kwargs
        )

    def get_outer_train(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the training set for risk assessment associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()

        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        return self._get_loader(
            train_indices, training_indices=train_indices, **kwargs
        )

    def get_outer_val(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the validation set for risk assessment associated with
        specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded.Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        val_indices = splitter.outer_folds[self.outer_k].val_idxs
        return self._get_loader(
            val_indices, training_indices=train_indices, **kwargs
        )

    def get_outer_test(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the test set for risk assessment associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        test_indices = splitter.outer_folds[self.outer_k].test_idxs
        return self._get_loader(
            test_indices, training_indices=train_indices, **kwargs
        )



from typing import List, Union, Tuple

import torch
from torch.nn import Module, CrossEntropyLoss, MSELoss, L1Loss

from mlwiz.util import s2c
from mlwiz.static import ARGS, CLASS_NAME
from mlwiz.training.event.handler import EventHandler
from mlwiz.training.event.state import State


class Metric(Module, EventHandler):
    r"""
    Metric is the main event handler for all metrics. Other metrics can easily
    subclass by implementing the :func:`forward` method, though sometimes
    more complex implementations are required.

    Args:
        use_as_loss (bool): whether this metric should act as a loss
            (i.e., it should act when :func:`on_backward` is called).
            **Used by MLWiz, no need to care about this.**
        reduction (str): the type of reduction to apply across samples of
            the mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        accumulate_over_epoch (bool): Whether or not to display the epoch-wise
            metric rather than an average of per-batch metrics.
            If true, it keep a list of predictions and target values across
            the entire epoch. Use it especially with batch-sensitive metrics,
            such as micro AP/F1 scores. Default is ``True``.
        force_cpu (bool): Whether or not to move all predictions to cpu
            before computing the epoch-wise loss/score. Default is ``True``.
        device (bool): The device used. Default is 'cpu'.
        kwargs (dict): additional arguments that may depend on the metric
    """

    def __init__(
        self,
        use_as_loss: bool = False,
        reduction: str = "mean",
        accumulate_over_epoch: bool = True,
        force_cpu: bool = True,
        device: str = "cpu",
        **kwargs: dict,
    ):
        super().__init__()
        self.batch_metrics = None
        self.use_as_loss = use_as_loss
        self.reduction = reduction
        self.accumulate_over_epoch = accumulate_over_epoch
        self.force_cpu = force_cpu
        self.device = device
        self.y_pred, self.y_true = None, None

        # Keeps track of how many times on_compute_metrics has been called
        # in the same batch.
        # Useful for temporal graph learning, for static graph learning it will
        # be always 1
        self.timesteps_in_batch = None

    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        raise NotImplementedError(
            "You should subclass Metric and implement this method!"
        )

    def get_main_metric_name(self) -> str:
        """
        Return the metric's main name.
        Useful when a metric is the combination of many.

        Returns:
            the metric's main name
        """
        return self.name

    def get_predictions_and_targets(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""
        Returns predictions and target tensors to be
        accumulated for a given metric

        Args:
            targets (:class:`torch.Tensor`): ground truth
            outputs (List[:class:`torch.Tensor`]): outputs of the model

        Returns:
            A tuple of tensors (predicted_values, target_values)
        """
        raise NotImplementedError(
            "You should subclass Metric and implement this method!"
        )

    def compute_metric(
        self, targets: torch.Tensor, predictions: torch.Tensor
    ) -> torch.tensor:
        r"""
        Computes the metric for a given set of targets and predictions

        Args:
            targets (:class:`torch.Tensor`): tensor of ground truth values
            predictions (:class:`torch.Tensor`):
                tensor of predictions of the model

        Returns:
            A tensor with the metric value
        """
        raise NotImplementedError(
            "You should subclass Metric and implement this method!"
        )

    def on_training_epoch_start(self, state: State):
        """
        Initialize list of batch metrics as well as the list of batch
        predictions and targets for the metric

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        """
        self.batch_metrics = []
        self.y_pred, self.y_true = {self.name: []}, {self.name: []}

    def on_training_batch_start(self, state: State):
        """
        Initializes the number of potential time steps in a batch
        (for temporal learning)

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        """
        self.timesteps_in_batch = 0.0

    def on_training_batch_end(self, state: State):
        """
        If we do not computed aggregated metric values over the entire epoch,
        populate the batch metrics list with the new loss/score.
        Divide by the number of timesteps in the batch
        (default is 1 for static datasets)

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        """
        if self.use_as_loss:
            batch_metric = state.batch_loss
        else:
            batch_metric = state.batch_score

        if not self.accumulate_over_epoch:
            self.batch_metrics.append(
                batch_metric[self.name].item() / self.timesteps_in_batch
            )

        self.timesteps_in_batch = None

    def on_training_epoch_end(self, state: State):
        """
        Computes the mean of batch metrics or an aggregated score over all
        epoch depending on the `accumulate_over_epoch` parameter. Updates
        `epoch_loss` and `epoch_score` fields in the state
        and resets the basic fields used.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        """
        if not self.accumulate_over_epoch:
            epoch_res = {
                self.name: torch.tensor(self.batch_metrics).sum()
                / len(self.batch_metrics)
            }
        else:
            epoch_res = {
                self.name: self.compute_metric(
                    torch.cat(self.y_true[self.name], dim=0),
                    torch.cat(self.y_pred[self.name], dim=0),
                )
            }

        if self.use_as_loss:
            state.update(epoch_loss=epoch_res)
        else:
            state.update(epoch_score=epoch_res)

        self.batch_metrics = None
        self.y_pred, self.y_true = None, None

    def on_eval_epoch_start(self, state: State):
        """
        Initialize list of batch metrics as well as the list of batch
        predictions and targets for the metric

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        """
        self.batch_metrics = []
        self.y_pred, self.y_true = {self.name: []}, {self.name: []}

    def on_eval_batch_start(self, state: State):
        """
        Initializes the number of potential time steps in a batch
        (for temporal learning)

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        """
        self.timesteps_in_batch = 0.0

    def on_eval_batch_end(self, state: State):
        """
        If we do not computed aggregated metric values over the entire epoch,
        populate the batch metrics list with the new loss/score.
        Divide by the number of timesteps in the batch
        (default is 1 for static datasets)

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        """
        if self.use_as_loss:
            batch_metric = state.batch_loss
        else:
            batch_metric = state.batch_score

        if not self.accumulate_over_epoch:
            self.batch_metrics.append(
                batch_metric[self.name].item() / self.timesteps_in_batch
            )

        self.timesteps_in_batch = None

    def on_eval_epoch_end(self, state: State):
        """
        Computes the mean of batch metrics or an aggregated score over all
        epoch depending on the `accumulate_over_epoch` parameter.
        Updates `epoch_loss` and `epoch_score` fields in the state
        and resets the basic fields used.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        """
        if not self.accumulate_over_epoch:
            epoch_res = {
                self.name: torch.tensor(self.batch_metrics).sum()
                / len(self.batch_metrics)
            }
        else:
            epoch_res = {
                self.name: self.compute_metric(
                    torch.cat(self.y_true[self.name], dim=0),
                    torch.cat(self.y_pred[self.name], dim=0),
                )
            }

        if self.use_as_loss:
            state.update(epoch_loss=epoch_res)
        else:
            state.update(epoch_score=epoch_res)

        self.batch_metrics = None
        self.y_pred, self.y_true = None, None

    def accumulate_predictions_and_targets(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> None:
        """
        Used to specify how to accumulate predictions and targets.
        This can be customized by subclasses like
        AdditiveLoss and MultiScore to accumulate predictions and
        targets for different losses/scores.

        Args:
            targets: target tensor
            *outputs: outputs of the model
        """
        y_pred_batch, y_true_batch = self.get_predictions_and_targets(
            targets, *outputs
        )
        metric_name = self.name

        self.y_pred[metric_name].append(
            y_pred_batch.detach().cpu()
            if self.force_cpu
            else y_pred_batch.detach()
        )
        self.y_true[metric_name].append(
            y_true_batch.detach().cpu()
            if self.force_cpu
            else y_true_batch.detach()
        )

    def on_compute_metrics(self, state: State):
        """
        Computes the loss/score depending on the metric, updating the
        `batch_loss` or `batch_score` field in the state.
        In temporal graph learning, this method is computed more than once
        before the batch ends, so we accumulate the loss or scores across
        timesteps of a single batch.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        """
        self.timesteps_in_batch += 1.0

        outputs, targets = state.batch_outputs, state.batch_targets

        # Loss case
        if self.use_as_loss:
            loss = self.forward(targets, *outputs)

            # [temporal graph learning] accumulate, rather than substitute,
            # the losses across multiple snapshots
            # these are reset accordingly by the Training Engine!
            if state.batch_loss is not None:
                old_loss = state.batch_loss
                loss = {k: old_loss[k] + loss[k] for k in loss.keys()}

            # this has to be updated per-batch no matter what
            state.update(batch_loss=loss)

        # Score case
        else:
            if not self.accumulate_over_epoch:
                score = self.forward(targets, *outputs)
                score = {
                    k: (
                        score[k].detach().cpu()
                        if self.force_cpu
                        else score[k].detach()
                    )
                    for k in score.keys()
                }

                # [temporal graph learning] accumulate, rather than substitute,
                # the scores across multiple snapshots
                # these are reset accordingly by the Training Engine!
                if state.batch_score is not None:
                    old_score = state.batch_score
                    score = {k: old_score[k] + score[k] for k in score.keys()}

                # mean of batches
                state.update(batch_score=score)

        if self.accumulate_over_epoch:
            self.accumulate_predictions_and_targets(targets, *outputs)

    def on_backward(self, state: State):
        """
        Calls backward on the loss if the metric is a loss.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        """
        if self.use_as_loss:
            try:
                state.batch_loss[self.name].backward()
            except Exception as e:
                # Here we catch potential multiprocessing related issues
                # see https://github.com/pytorch/pytorch/wiki/Autograd-and-Fork
                raise (e)

    def forward(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> dict:
        r"""
        Computes the metric value. Optionally, and only for scores used
        as losses, some extra information can be also returned.

        Args:
            targets (:class:`torch.Tensor`): ground truth
            outputs (List[:class:`torch.Tensor`]): outputs of the model
            batch_loss_extra (dict): dictionary of information
                computed by metrics used as losses

        Returns:
            A dictionary containing associations metric_name - value
        """
        y_pred_batch, y_true_batch = self.get_predictions_and_targets(
            targets, *outputs
        )
        return {self.name: self.compute_metric(y_true_batch, y_pred_batch)}


class MultiScore(Metric):
    r"""
    This class is used to keep track of multiple additional metrics
    used as scores, rather than losses.

    Args:
        use_as_loss (bool): whether this metric should act as a loss
            (i.e., it should act when :func:`on_backward` is called).
            **Used by MLWiz, no need to care about this.**
        reduction (str): the type of reduction to apply across samples of the
            mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        accumulate_over_epoch (bool): Whether or not to display the epoch-wise
            metric rather than an average of per-batch metrics.  If true, it
            keeps a list of predictions and target values across the entire
            epoch. Use it especially with batch-sensitive metrics,
            such as micro AP/F1 scores. Default is ``True``.
        force_cpu (bool): Whether or not to move all predictions to cpu
            before computing the epoch-wise loss/score. Default is ``True``.
        device (bool): The device used. Default is 'cpu'.
        main_scorer (:class:`~mlwiz.training.callback.metric.Metric`): the
            score on which final results are computed.
        extra_scorers (dict): dictionary of other metrics to consider.
    """

    def __init__(
        self,
        use_as_loss,
        reduction="mean",
        accumulate_over_epoch: bool = True,
        force_cpu: bool = True,
        device: str = "cpu",
        main_scorer=None,
        **extra_scorers,
    ):
        assert not use_as_loss, "MultiScore cannot be used as loss"
        assert main_scorer is not None, "You have to provide a main scorer"
        super().__init__(
            use_as_loss, reduction, accumulate_over_epoch, force_cpu, device
        )
        self.scores = [self._istantiate_scorer(main_scorer)] + [
            self._istantiate_scorer(score) for score in extra_scorers.values()
        ]

    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        return "Multi Score"

    def _istantiate_scorer(self, scorer):
        """
        Istantiate a scorer with its own arguments (if any are given)
        """
        if isinstance(scorer, dict):
            args = scorer[ARGS]
            return s2c(scorer[CLASS_NAME])(use_as_loss=False, **args)
        else:
            return s2c(scorer)(use_as_loss=False)

    def get_main_metric_name(self):
        """
        Returns the name of the first scorer that is passed to this class via
        the `__init__` method.
        """
        return self.scores[0].get_main_metric_name()

    def on_training_epoch_start(self, state: State):
        """
        Compared to superclass version, initializes a dictionary for each score
        to track rather than single lists

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        self.batch_metrics = {s.name: [] for s in self.scores}
        self.y_pred = {s.name: [] for s in self.scores}
        self.y_true = {s.name: [] for s in self.scores}

    def on_training_batch_end(self, state: State):
        """
        For each scorer, computes the average metric in the batch wrt the
        number of timesteps (default is 1 for static datasets) unless
        statistics are accumulated over the entire epoch

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information

        """
        if not self.accumulate_over_epoch:
            for k, v in state.batch_score.items():
                self.batch_metrics[k].append(
                    v.item() / self.timesteps_in_batch
                )

        self.timesteps_in_batch = None

    def on_training_epoch_end(self, state: State):
        """
        For each score, computes the epoch scores using the same
        ogic as the superclass

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if not self.accumulate_over_epoch:
            epoch_res = {
                s.name: torch.tensor(self.batch_metrics[s.name]).sum()
                / len(self.batch_metrics[s.name])
                for s in self.scores
            }
        else:
            epoch_res = {
                s.name: s.compute_metric(
                    torch.cat(self.y_true[s.name], dim=0),
                    torch.cat(self.y_pred[s.name], dim=0),
                )
                for s in self.scores
            }

        state.update(epoch_score=epoch_res)

        self.batch_metrics = None
        self.y_pred, self.y_true = None, None

    def on_eval_epoch_start(self, state: State):
        """
        Compared to superclass version, initializes a dictionary for each
        score to track rather than single lists

         Args:
             state (:class:`~training.event.state.State`):
                object holding training information
        """
        self.batch_metrics = {s.name: [] for s in self.scores}
        self.y_pred = {s.name: [] for s in self.scores}
        self.y_true = {s.name: [] for s in self.scores}

    def on_eval_batch_end(self, state: State):
        """
        For each scorer, computes the average metric in the batch wrt the
        number of timesteps (default is 1 for static datasets) unless
        statistics are accumulated over the entire epoch

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if not self.accumulate_over_epoch:
            for k, v in state.batch_score.items():
                self.batch_metrics[k].append(
                    v.item() / self.timesteps_in_batch
                )

        self.timesteps_in_batch = None

    def on_eval_epoch_end(self, state: State):
        """
        For each score, computes the epoch scores using the
        same logic as the superclass

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if not self.accumulate_over_epoch:
            epoch_res = {
                s.name: torch.tensor(self.batch_metrics[s.name]).sum()
                / len(self.batch_metrics[s.name])
                for s in self.scores
            }
        else:
            epoch_res = {
                s.name: s.compute_metric(
                    torch.cat(self.y_true[s.name], dim=0),
                    torch.cat(self.y_pred[s.name], dim=0),
                )
                for s in self.scores
            }

        state.update(epoch_score=epoch_res)

        self.batch_metrics = None
        self.y_pred, self.y_true = None, None

    def accumulate_predictions_and_targets(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> None:
        """
        Accumulates predictions and targets of the batch into a list for
        each scorer, so as to compute aggregated statistics at the
        end of an epoch.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        for s in self.scores:
            y_pred_batch, y_true_batch = s.get_predictions_and_targets(
                targets, *outputs
            )
            metric_name = s.name

            self.y_pred[metric_name].append(
                y_pred_batch.detach().cpu()
                if self.force_cpu
                else y_pred_batch.detach()
            )
            self.y_true[metric_name].append(
                y_true_batch.detach().cpu()
                if self.force_cpu
                else y_true_batch.detach()
            )

    def forward(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> Union[dict, float]:
        """
        For each scorer, it computes a score and returns them into a dictionary

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        res = {}
        for s in self.scores:
            y_pred_batch, y_true_batch = s.get_predictions_and_targets(
                targets, *outputs
            )
            res.update({s.name: s.compute_metric(y_true_batch, y_pred_batch)})
        return res


class AdditiveLoss(Metric):
    r"""
    AdditiveLoss sums an arbitrary number of losses together.

    Args:
        use_as_loss (bool): whether this metric should act as a loss
            (i.e., it should act when :func:`on_backward` is called).
            **Used by MLWiz, no need to care about this.**
        reduction (str): the type of reduction to apply across samples of the
            mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        accumulate_over_epoch (bool): Whether or not to display the epoch-wise
            metric rather than an average of per-batch metrics. If true,
            it keeps a list of predictions and target values across the
            entire epoch. Use it especially with batch-sensitive metrics,
            such as micro AP/F1 scores. Default is ``True``.
        force_cpu (bool): Whether or not to move all predictions to cpu
            before computing the epoch-wise loss/score. Default is ``True``.
        device (bool): The device used. Default is 'cpu'.
        losses_weights: (dict): dictionary of (loss_name, loss_weight) that
            specifies the weight to apply to each loss to be summed.
        losses (dict): dictionary of metrics to add together
    """

    def __init__(
        self,
        use_as_loss,
        reduction="mean",
        accumulate_over_epoch: bool = True,
        force_cpu: bool = True,
        device: str = "cpu",
        losses_weights: dict = None,
        **losses: dict,
    ):
        assert use_as_loss, "Additive loss can only be used as a loss"
        super().__init__(
            use_as_loss, reduction, accumulate_over_epoch, force_cpu, device
        )

        self.losses_weights = losses_weights
        self.losses = [
            self._instantiate_loss(loss) for loss in losses.values()
        ]

        if self.losses_weights is not None:
            for loss in self.losses:
                # check that a weight exists for each loss
                assert loss.name in self.losses_weights, (
                    "You have to specify a weight for each loss! "
                    f"We could not find the weight for {loss.name} "
                    f"in the dict."
                )
        else:
            # all losses are simply added together
            self.losses_weights = {loss.name: 1.0 for loss in self.losses}

    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        return "Additive Loss"

    def _instantiate_loss(self, loss):
        """
        Istantiate a loss with its own arguments (if any are given)
        """
        if isinstance(loss, dict):
            args = loss[ARGS]
            return s2c(loss[CLASS_NAME])(use_as_loss=True, **args)
        else:
            return s2c(loss)(use_as_loss=True)

    def on_training_epoch_start(self, state: State):
        """
        Instantiates a dictionary with one list per loss (including itself,
        representing the sum of all losses)

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        self.batch_metrics = {loss.name: [] for loss in [self] + self.losses}
        self.y_pred = {loss.name: [] for loss in [self] + self.losses}
        self.y_true = {loss.name: [] for loss in [self] + self.losses}

    def on_training_batch_end(self, state: State):
        """
        For each loss, computes the average metric in the batch wrt the number
        of timesteps (default is 1 for static datasets)
        unless statistics are accumulated over the entire epoch

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if not self.accumulate_over_epoch:
            for k, v in state.batch_loss.items():
                self.batch_metrics[k].append(
                    v.item() / self.timesteps_in_batch
                )

        self.timesteps_in_batch = None

    def on_training_epoch_end(self, state: State):
        """
        Computes an averaged or aggregated loss across the entire epoch,
        including itself as the main loss. Updates the field `epoch_loss`
        in state.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if not self.accumulate_over_epoch:
            epoch_res = {
                loss.name: torch.tensor(self.batch_metrics[loss.name]).sum()
                / len(self.batch_metrics[loss.name])
                for loss in [self] + self.losses
            }
        else:
            epoch_res = {
                loss.name: loss.compute_metric(
                    torch.cat(self.y_true[loss.name], dim=0),
                    torch.cat(self.y_pred[loss.name], dim=0),
                )
                * self.losses_weights[loss.name]
                for loss in self.losses
            }
            additive_epoch_loss = 0.0
            for loss in self.losses:
                additive_epoch_loss += epoch_res[loss.name]
            epoch_res[self.name] = additive_epoch_loss

        state.update(epoch_loss=epoch_res)

        self.batch_metrics = None
        self.y_pred, self.y_true = None, None

    def on_eval_epoch_start(self, state: State):
        """
        Instantiates a dictionary with one list per loss (including itself,
        representing the sum of all losses)

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        self.batch_metrics = {loss.name: [] for loss in [self] + self.losses}
        self.y_pred = {loss.name: [] for loss in [self] + self.losses}
        self.y_true = {loss.name: [] for loss in [self] + self.losses}

    def on_eval_epoch_end(self, state: State):
        """
        Computes an averaged or aggregated loss across the entire epoch,
        including itself as the main loss. Updates the field `epoch_loss`
        in state.

        Args:
            state (:class:`~training.event.state.State`):
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if not self.accumulate_over_epoch:
            epoch_res = {
                loss.name: torch.tensor(self.batch_metrics[loss.name]).sum()
                / len(self.batch_metrics[loss.name])
                for loss in [self] + self.losses
            }
        else:
            epoch_res = {
                loss.name: loss.compute_metric(
                    torch.cat(self.y_true[loss.name], dim=0),
                    torch.cat(self.y_pred[loss.name], dim=0),
                )
                * self.losses_weights[loss.name]
                for loss in self.losses
            }
            additive_epoch_loss = 0.0
            for loss in self.losses:
                additive_epoch_loss += epoch_res[loss.name]
            epoch_res[self.name] = additive_epoch_loss

        state.update(epoch_loss=epoch_res)

        self.batch_metrics = None
        self.y_pred, self.y_true = None, None

    def on_eval_batch_end(self, state: State):
        """
        For each loss, computes the average metric in the batch wrt the number
        of timesteps (default is 1 for static datasets)
        unless statistics are accumulated over the entire epoch

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if not self.accumulate_over_epoch:
            for k, v in state.batch_loss.items():
                self.batch_metrics[k].append(
                    v.item() / self.timesteps_in_batch
                )

        self.timesteps_in_batch = None

    def compute_metric(
        self, targets: torch.Tensor, predictions: torch.Tensor
    ) -> torch.tensor:
        """
        Sums the value of all different losses into one

        Args:
            targets (:class:`torch.Tensor`): tensor of ground truth values
            predictions (:class:`torch.Tensor`):
                tensor of predictions of the model

        Returns:
            A tensor with the metric value
        """
        loss_sum = 0.0
        for loss in self.losses:
            single_loss = (
                loss.compute_metric(targets, predictions)
                * self.losses_weights[loss.name]
            )
            loss_sum += single_loss
        return loss_sum

    def accumulate_predictions_and_targets(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> None:
        """
        Accumulates predictions and targets of the batch into a list
        for each loss, so as to compute aggregated statistics at the end
        of an epoch.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        for loss in self.losses:
            y_pred_batch, y_true_batch = loss.get_predictions_and_targets(
                targets, *outputs
            )
            metric_name = loss.name

            self.y_pred[metric_name].append(
                y_pred_batch.detach().cpu()
                if self.force_cpu
                else y_pred_batch.detach()
            )
            self.y_true[metric_name].append(
                y_true_batch.detach().cpu()
                if self.force_cpu
                else y_true_batch.detach()
            )

    def forward(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> dict:
        """
        For each scorer, it computes a loss and returns them into a dictionary,
        alongside the sum of all losses.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        res = {}
        loss_sum = 0.0

        for loss in self.losses:
            y_pred_batch, y_true_batch = loss.get_predictions_and_targets(
                targets, *outputs
            )
            single_loss = (
                loss.compute_metric(y_true_batch, y_pred_batch)
                * self.losses_weights[loss.name]
            )

            res[loss.name] = single_loss
            loss_sum += single_loss

        res[self.name] = loss_sum
        return res


class Classification(Metric):
    r"""
    Generic metric for classification tasks. Used to maximize code reuse
    for classical metrics.
    """

    def __init__(
        self,
        use_as_loss=False,
        reduction="mean",
        accumulate_over_epoch: bool = True,
        force_cpu: bool = True,
        device: str = "cpu",
        **kwargs: dict,
    ):
        super().__init__(
            use_as_loss=use_as_loss,
            reduction=reduction,
            accumulate_over_epoch=accumulate_over_epoch,
            force_cpu=force_cpu,
            device=device,
            **kwargs,
        )
        self.metric = None

    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        return "Classification Metric"

    def get_predictions_and_targets(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Returns output[0] as predictions and dataset targets. Squeezes the
        first dimension of output and targets to get single vector.

        Args:
            targets (:class:`torch.Tensor`): ground truth
            outputs (List[:class:`torch.Tensor`]): outputs of the model

        Returns:
            A tuple of tensors (predicted_values, target_values)
        """
        outputs = outputs[0].squeeze(dim=1)

        if len(targets.shape) == 2:
            targets = targets.squeeze(dim=1)

        return outputs, targets

    def compute_metric(
        self, targets: torch.Tensor, predictions: torch.Tensor
    ) -> torch.tensor:
        """
        Applies a classification metric
        (to be subclassed as it is None in this class)

        Args:
            targets (:class:`torch.Tensor`): tensor of ground truth values
            predictions (:class:`torch.Tensor`):
                tensor of predictions of the model

        Returns:
            A tensor with the metric value
        """
        metric = self.metric(predictions, targets)
        return metric


class Regression(Metric):
    r"""
    Generic metric for regression tasks. Used to maximize code reuse
    for classical metrics.
    """

    def __init__(
        self,
        use_as_loss=False,
        reduction="mean",
        accumulate_over_epoch: bool = True,
        force_cpu: bool = True,
        device: str = "cpu",
        **kwargs: dict,
    ):
        super().__init__(
            use_as_loss=use_as_loss,
            reduction=reduction,
            accumulate_over_epoch=accumulate_over_epoch,
            force_cpu=force_cpu,
            device=device,
            **kwargs,
        )
        self.metric = None

    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        return "Regression Metric"

    def get_predictions_and_targets(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Returns output[0] as predictions and dataset targets.
        Squeezes the first dimension of output and targets to get
        single vector.

        Args:
            targets (:class:`torch.Tensor`): ground truth
            outputs (List[:class:`torch.Tensor`]): outputs of the model

        Returns:
            A tuple of tensors (predicted_values, target_values)
        """
        outputs = outputs[0].squeeze(dim=1)

        if len(targets.shape) == 2:
            targets = targets.squeeze(dim=1)

        return outputs, targets

    def compute_metric(
        self, targets: torch.Tensor, predictions: torch.Tensor
    ) -> torch.tensor:
        """
        Applies a regression metric
        (to be subclassed as it is None in this class)

        Args:
            targets (:class:`torch.Tensor`): tensor of ground truth values
            predictions (:class:`torch.Tensor`):
                tensor of predictions of the model

        Returns:
            A tensor with the metric value
        """
        metric = self.metric(predictions, targets)
        return metric


class MulticlassClassification(Classification):
    r"""
    Wrapper around :class:`torch.nn.CrossEntropyLoss`
    """

    def __init__(
        self,
        use_as_loss=False,
        reduction="mean",
        accumulate_over_epoch: bool = True,
        force_cpu: bool = True,
        device: str = "cpu",
        **kwargs: dict,
    ):
        super().__init__(
            use_as_loss=use_as_loss,
            reduction=reduction,
            accumulate_over_epoch=accumulate_over_epoch,
            force_cpu=force_cpu,
            device=device,
            **kwargs,
        )
        self.metric = CrossEntropyLoss(reduction=reduction)

    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        return "Multiclass Classification"


class MeanSquareError(Regression):
    r"""
    Wrapper around :class:`torch.nn.MSELoss`
    """

    def __init__(
        self,
        use_as_loss=False,
        reduction="mean",
        accumulate_over_epoch: bool = True,
        force_cpu: bool = True,
        device: str = "cpu",
        **kwargs: dict,
    ):
        super().__init__(
            use_as_loss=use_as_loss,
            reduction=reduction,
            accumulate_over_epoch=accumulate_over_epoch,
            force_cpu=force_cpu,
            device=device,
            **kwargs,
        )
        self.metric = MSELoss(reduction=reduction)

    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        return "Mean Square Error"


class MeanAverageError(Regression):
    r"""
    Wrapper around :class:`torch.nn.MSELoss`
    """

    def __init__(
        self,
        use_as_loss=False,
        reduction="mean",
        accumulate_over_epoch: bool = True,
        force_cpu: bool = True,
        device: str = "cpu",
        **kwargs: dict,
    ):
        super().__init__(
            use_as_loss=use_as_loss,
            reduction=reduction,
            accumulate_over_epoch=accumulate_over_epoch,
            force_cpu=force_cpu,
            device=device,
            **kwargs,
        )
        self.metric = L1Loss(reduction=reduction)

    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        return "Mean Average Error"


class MulticlassAccuracy(Metric):
    """
    Implements multiclass classification accuracy.
    """

    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        return "Multiclass Accuracy"

    @staticmethod
    def _get_correct(output):
        """
        Returns the argmax of the output alongside dimension 1.
        """
        return torch.argmax(output, dim=1)

    def get_predictions_and_targets(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Takes output[0] as predictions and computes a discrete class
        using argmax. Returns standard dataset targets as well. Squeezes
        the first dimension of output and targets to get single vector.

        Args:
            targets (:class:`torch.Tensor`): ground truth
            outputs (List[:class:`torch.Tensor`]): outputs of the model

        Returns:
            A tuple of tensors (predicted_values, target_values)
        """
        pred = outputs[0]
        correct = self._get_correct(pred)

        if len(targets.shape) == 2:
            targets = targets.squeeze(dim=1)

        return correct, targets

    def compute_metric(
        self, targets: torch.Tensor, predictions: torch.Tensor
    ) -> torch.tensor:
        """
        Takes output[0] as predictions and computes a discrete class using
        argmax. Returns standard dataset targets as well. Squeezes the first
        dimension of output and targets to get single vector.

        Args:
            targets (:class:`torch.Tensor`): tensor of ground truth values
            predictions (:class:`torch.Tensor`):
                tensor of predictions of the model

        Returns:
            A tensor with the metric value
        """
        metric = (
            100.0 * (predictions == targets).sum().float() / targets.size(0)
        )
        return metric


class AllocatedGPUMemory(Metric):
    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        return "GPU Memory Allocated (MBs)"

    def get_predictions_and_targets(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        gpu_mem = torch.tensor([torch.cuda.memory_allocated()]).float() / (
            1024 * 1024
        )
        return gpu_mem, gpu_mem

    def compute_metric(
        self, targets: torch.Tensor, predictions: torch.Tensor
    ) -> torch.tensor:
        return predictions.mean()


class ToyMetric(Metric):
    r"""
    Implements a toy metric.
    """

    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        return "Toy Metric"

    @staticmethod
    def _get_correct(output):
        """
        Returns the argmax of the output alongside dimension 1.
        """
        return torch.argmax(output, dim=1)

    def get_predictions_and_targets(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Returns output[0] and dataset targets

        Args:
            targets (:class:`torch.Tensor`): ground truth
            outputs (List[:class:`torch.Tensor`]): outputs of the model

        Returns:
            A tuple of tensors (predicted_values, target_values)
        """
        return outputs[0], targets

    def compute_metric(
        self, targets: torch.Tensor, predictions: torch.Tensor
    ) -> torch.tensor:
        """
        Computes a dummy score

        Args:
            targets (:class:`torch.Tensor`): tensor of ground truth values
            predictions (:class:`torch.Tensor`):
                tensor of predictions of the model

        Returns:
            A tensor with the metric value
        """
        metric = 100.0 * torch.ones(1)
        return metric


class SingleGraphMulticlassClassification(MulticlassClassification):
    r"""
    Wrapper around :class:`torch.nn.CrossEntropyLoss`
    """

    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        return "Single Graph Multiclass Classification"

    def get_predictions_and_targets(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        # o and h have alredy been filtered by the model according to
        # required data split indexes
        o, _, idxs = outputs
        t = targets[idxs]
        return o, t


class SingleGraphMulticlassAccuracy(MulticlassAccuracy):
    r"""
    Wrapper around :class:`torch.nn.CrossEntropyLoss`
    """

    @property
    def name(self) -> str:
        """
        The name of the loss to be used in configuration files and displayed
        on Tensorboard
        """
        return "Single Graph Multiclass Accuracy"

    def get_predictions_and_targets(
        self, targets: torch.Tensor, *outputs: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        # o and h have alredy been filtered by the model according to
        # required data split indexes
        o, _, idxs = outputs
        pred = self._get_correct(o)
        if len(targets.shape) == 2:
            targets = targets.squeeze(dim=1)
        t = targets[idxs]

        return pred, t



import operator
from copy import deepcopy
from pathlib import Path

from mlwiz.static import *
from mlwiz.training.event.handler import EventHandler
from mlwiz.training.event.state import State
from mlwiz.training.util import atomic_save


class EarlyStopper(EventHandler):
    """
    EarlyStopper is the main event handler for optimizers. Just create a
    subclass that implements an early stopping method.

    Args:
        monitor (str): the metric to monitor. The format is
            ``[TRAINING|VALIDATION]_[METRIC NAME]``, where
            ``TRAINING`` and ``VALIDATION`` are defined in ``mlwiz.static``
        mode (str): can be ``MIN``
            or ``MAX`` (as defined in ``mlwiz.static``)
        checkpoint (bool): whether we are interested in the
            checkpoint of the "best" epoch or not
    """

    def __init__(self, monitor: str, mode: str, checkpoint: bool = False):
        super().__init__()
        self.monitor = monitor
        self.best_metric = None
        self.checkpoint = checkpoint

        if MIN in mode:
            self.operator = operator.le
        elif MAX in mode:
            self.operator = operator.ge
        else:
            raise NotImplementedError("Mode not understood by early stopper.")

        assert (
            TEST not in monitor
        ), "Do not apply early stopping to the test set!"

    def on_epoch_end(self, state: State):
        """
        At the end of an epoch, check that the validation score improves over
        the current best validation score. If so, store the necessary info
        in a dictionary and save it into the "best_epoch_results" property
        of the state. If it is time to stop, updates the
        `stop_training` field of the state.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        # it is possible that we evaluate every `n` epochs
        if not (
            self.monitor in state.epoch_results[SCORES]
            or self.monitor in state.epoch_results[LOSSES]
        ):
            return

        if self.monitor in state.epoch_results[SCORES]:
            score_or_loss = SCORES
        else:
            score_or_loss = LOSSES
        metric_to_compare = state.epoch_results[score_or_loss][self.monitor]

        if not hasattr(state, BEST_EPOCH_RESULTS):
            state.update(best_epoch_results=deepcopy(state.epoch_results))
            state.best_epoch_results[BEST_EPOCH] = state.epoch
            state.best_epoch_results[score_or_loss][
                self.monitor
            ] = state.epoch_results[score_or_loss][self.monitor]
            state.best_epoch_results[MODEL_STATE] = deepcopy(
                state.model.state_dict()
            )
            state.best_epoch_results[OPTIMIZER_STATE] = state[
                OPTIMIZER_STATE
            ]  # computed by optimizer
            state.best_epoch_results[SCHEDULER_STATE] = state[
                SCHEDULER_STATE
            ]  # computed by scheduler
            if self.checkpoint:
                atomic_save(
                    state.best_epoch_results,
                    Path(state.exp_path, BEST_CHECKPOINT_FILENAME),
                )
        else:
            best_metric = state.best_epoch_results[score_or_loss][self.monitor]

            if self.operator(metric_to_compare, best_metric):
                state.update(best_epoch_results=deepcopy(state.epoch_results))
                state.best_epoch_results[BEST_EPOCH] = state.epoch
                state.best_epoch_results[score_or_loss][
                    self.monitor
                ] = metric_to_compare
                state.best_epoch_results[MODEL_STATE] = deepcopy(
                    state.model.state_dict()
                )
                state.best_epoch_results[OPTIMIZER_STATE] = state[
                    OPTIMIZER_STATE
                ]  # computed by optimizer
                state.best_epoch_results[SCHEDULER_STATE] = state[
                    SCHEDULER_STATE
                ]  # computed by scheduler
                if self.checkpoint:
                    atomic_save(
                        state.best_epoch_results,
                        Path(state.exp_path, BEST_CHECKPOINT_FILENAME),
                    )

        # Regarless of improvement or not
        stop_training = self.stop(state, score_or_loss, metric_to_compare)
        state.update(stop_training=stop_training)

    def stop(self, state: State, score_or_loss: str, metric: str) -> bool:
        """
        Returns true when the early stopping technique decides
        it is time to stop.

        Args:
            state (:class:`~training.event.state.State`): object holding
                training information
            score_or_loss (str): whether to monitor scores or losses
            metric (str): the metric to consider. The format is
                ``[TRAINING|VALIDATION]_[METRIC NAME]``, where
                ``TRAINING`` and ``VALIDATION`` are defined in ``mlwiz.static``

        Returns:
            a boolean specifying whether training should be stopped or not
        """
        raise NotImplementedError(
            "Sublass EarlyStopper and implement this method!"
        )


class PatienceEarlyStopper(EarlyStopper):
    """
    Early Stopper that implements patience

    Args:
        monitor (str): the metric to monitor. The format is
            ``[TRAINING|VALIDATION]_[METRIC NAME]``, where
            ``TRAINING`` and ``VALIDATION`` are defined in ``mlwiz.static``
        mode (str): can be ``MIN`` or ``MAX``
            (as defined in ``mlwiz.static``)
        patience (int): the number of epochs of patience
        checkpoint (bool): whether we are interested in the checkpoint
            of the "best" epoch or not
    """

    def __init__(self, monitor, mode, patience=30, checkpoint=False):
        super().__init__(monitor, mode, checkpoint)
        self.patience = patience

    def stop(self, state, score_or_loss, metric):
        """
        Returns true when the number of epochs without improvement
        is greater than our patience parameter.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
            score_or_loss (str): whether to monitor scores or losses
            metric (str): the metric to consider. The format is
                ``[TRAINING|VALIDATION]_[METRIC NAME]``, where
                ``TRAINING`` and ``VALIDATION`` are defined in ``mlwiz.static``

        Returns:
            a boolean specifying whether training should be stopped or not
        """
        # do not start with patience until you have evaluated at least once
        if not hasattr(state, BEST_EPOCH_RESULTS):
            return False

        best_epoch = state.best_epoch_results[BEST_EPOCH]
        stop_training = (state.epoch - best_epoch) >= self.patience
        return stop_training



from typing import List, Optional, Tuple, Union

import torch
from torch.nn import Linear

from mlwiz.model.interface import ModelInterface


class GRU(ModelInterface):
    """
    An RNN model used to test the library.
    """

    def __init__(
        self,
        dim_input_features: Union[int, Tuple[int]],
        dim_target: int,
        config: dict,
    ):
        super().__init__(
            dim_input_features,
            dim_target,
            config,
        )
        assert type(dim_input_features) == int

        dim_embedding = config["dim_embedding"]

        self.rnn = torch.nn.GRU(
            input_size=dim_input_features,
            hidden_size=dim_embedding,
            num_layers=1,
            batch_first=True,
        )
        self.out_layer = Linear(dim_embedding, dim_target)

    def forward(
        self, data: torch.Tensor
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        """
        Implements an MLP forward pass

        Args:
            data (torch.Tensor): a batched tensor

        Returns:
            a tuple (output, node_embedddings)
        """
        h, hidden_state = self.rnn(data)
        # get last time step
        h_last = h[:, -1, :]  # batch size x dim_embedding
        o = self.out_layer(h_last)
        return o, h
